from torch import nn
from torch.nn.init import xavier_uniform_
from torch.nn.init import zeros_

from newtonnet.layers.batchrenorm import BatchRenorm1d


class Dense(nn.Linear):
    r"""
    Fully connected linear layer with activation function.

    credit : https://github.com/atomistic-machine-learning/schnetpack under the MIT License

    .. math::
        x = normalize(x)
        y = activation(xW^T + b)
        y = dropout(y)

    Parameters
    ----------
    in_features: int
        number of input feature :math:`x`.
    out_features: int
        number of output features :math:`y`.
    bias: bool, optional (default: True)
        if False, the layer will not adapt bias :math:`b`.
    activation: callable, optional (default: None)
        A torch.nn activation function (e.g., torch.nn.ReLU())
        if None, no activation function is used.
    weight_init: callable, optional (default: torch.nn.init.xavier_uniform_)
        weight initializer
    bias_init: callable, optional (default: torch.nn.init.zeros_)
        bias initializer
    dropout: float, optional (default: None)
        if the value is a float, it will be used as dropout probabilty
    norm: bool, optional (default: None)
        if True, batch normalization will be applied prior to the Linear mapping.


    """
    def __init__(self,
                 in_features,
                 out_features,
                 bias=True,
                 activation=None,
                 weight_init=xavier_uniform_,
                 xavier_init_gain=1.,
                 bias_init=zeros_,
                 dropout=None,
                 norm=None):
        self.weight_init = weight_init
        self.gain = xavier_init_gain
        self.bias_init = bias_init
        super(Dense, self).__init__(in_features, out_features, bias)
        self.activation = activation
        # initialize linear layer y = xW^T + b

        self.dropout = dropout
        if dropout:
            self.dropout = nn.Dropout(dropout)

        self.norm = norm
        if norm:
            self.norm = BatchRenorm1d(
                num_features=out_features
                # track_running_stats=False
            )  #, momentum=0.99, eps=0.001) # momentum and eps are based on Keras default values

    def reset_parameters(self):
        self.weight_init(self.weight, gain=self.gain)
        if self.bias is not None:
            self.bias_init(self.bias)

    def forward(self, x):
        """

        Parameters
        ----------
        x: torch.Tensor
            A tensor with shape (batch_size, in_features)


        Returns
        -------
        torch.Tensor
            A tensor with shape (batch_size, out_features)


        """
        # compute linear layer y = xW^T + b
        x = super(Dense, self).forward(x)

        # batch normalization
        if self.norm:
            x = self.norm(x)

        # add activation function
        if self.activation:
            x = self.activation(x)

        # dropout
        if self.dropout:
            x = self.dropout(x)

        return x



# Written by Roman Zubatyuk and Justin S. Smith
import h5py
import numpy as np
import platform
import os


PY_VERSION = int(platform.python_version().split('.')[0]) > 3

class datapacker(object):
    def __init__(self, store_file, mode='w-', complib='gzip', complevel=6):
        """Wrapper to store arrays within HFD5 file
        """
        # opening file
        self.store = h5py.File(store_file, mode=mode)
        self.clib = complib
        self.clev = complevel

    def store_data(self, store_loc, **kwargs):
        """Put arrays to store
        """
        #print(store_loc)
        g = self.store.create_group(store_loc)
        for k, v, in kwargs.items():
            #print(type(v[0]))

            #print(k)
            if type(v) == list:
                if len(v) != 0:
                    if type(v[0]) is np.str_ or type(v[0]) is str:
                        v = [a.encode('utf8') for a in v]

            g.create_dataset(k, data=v, compression=self.clib, compression_opts=self.clev)

    def cleanup(self):
        """Wrapper to close HDF5 file
        """
        self.store.close()


class anidataloader(object):

    ''' Contructor '''
    def __init__(self, store_file):
        if not os.path.exists(store_file):
            exit('Error: file not found - '+store_file)
        self.store = h5py.File(store_file)

    ''' Group recursive iterator (iterate through all groups in all branches and return datasets in dicts) '''
    def h5py_dataset_iterator(self,g, prefix=''):
        for key in g.keys():
            item = g[key]
            path = '{}/{}'.format(prefix, key)
            keys = [i for i in item.keys()]
            if isinstance(item[keys[0]], h5py.Dataset): # test for dataset
                data = {'path':path}
                for k in keys:
                    if not isinstance(item[k], h5py.Group):
                        if int(h5py.__version__.split('.')[0]) >= 3: # check h5py version
                            dataset = item[k][()]
                        else:
                            dataset = np.array(item[k].value)

                        if type(dataset) is np.ndarray:
                            if dataset.size != 0:
                                if type(dataset[0]) is np.bytes_:
                                    dataset = [a.decode('ascii') for a in dataset]

                        data.update({k:dataset})

                yield data
            else: # test for group (go down)
                yield from self.h5py_dataset_iterator(item, path)

    ''' Default class iterator (iterate through all data) '''
    def __iter__(self):
        for data in self.h5py_dataset_iterator(self.store):
            yield data

    ''' Returns a list of all groups in the file '''
    def get_group_list(self):
        return [g for g in self.store.values()]

    ''' Allows interation through the data in a given group '''
    def iter_group(self,g):
        for data in self.h5py_dataset_iterator(g):
            yield data

    ''' Returns the requested dataset '''
    def get_data(self, path, prefix=''):
        item = self.store[path]
        path = '{}/{}'.format(prefix, path)
        keys = [i for i in item.keys()]
        data = {'path': path}
        # print(path)
        for k in keys:
            if not isinstance(item[k], h5py.Group):
                if int(h5py.__version__.split('.')[0]) >= 3: # check h5py version
                    dataset = item[k][()]
                else:
                    dataset = np.array(item[k].value)

                if type(dataset) is np.ndarray:
                    if dataset.size != 0:
                        if type(dataset[0]) is np.bytes_:
                            dataset = [a.decode('ascii') for a in dataset]

                data.update({k: dataset})
        return data

    ''' Returns the number of groups '''
    def group_size(self):
        return len(self.get_group_list())

    def size(self):
        count = 0
        for g in self.store.values():
            count = count + len(g.items())
        return count

    ''' Close the HDF5 file '''
    def cleanup(self):
        self.store.close()




import numpy as np


def euler_rotation_matrix(theta):
    """
    Rotate the xyz values based on the euler angles, theta. Directly copied from:
    Credit : "https://www.learnopencv.com/rotation-matrix-to-euler-angles/"

    Parameters
    ----------
    theta: numpy array
        A 1D array of angles along x, y and z directions

    Returns
    -------
    numpy array: rotation matrix with shape (3,3)
    """

    R_x = np.array([[1, 0, 0], [0, np.cos(theta[0]), -np.sin(theta[0])],
                    [0, np.sin(theta[0]),
                     np.cos(theta[0])]])
    R_y = np.array([[np.cos(theta[1]), 0,
                     np.sin(theta[1])], [0, 1, 0],
                    [-np.sin(theta[1]), 0,
                     np.cos(theta[1])]])
    R_z = np.array([[np.cos(theta[2]), -np.sin(theta[2]), 0],
                    [np.sin(theta[2]), np.cos(theta[2]), 0], [0, 0, 1]])
    R = np.dot(R_z, np.dot(R_y, R_x))

    return R


def rotate_molecule(atoms, theta=None):
    """
    Rotates the structure of molecule between -pi/2 and pi/2.

    Parameters
    ----------
    atoms: numpy array
        An array of atomic positions with last dimension = 3

    theta: numpy array, optional (default: None)
        A 1D array of angles along x, y and z directions.
        If None, it will be generated uniformly between -pi/2 and pi/2

    Returns
    -------
    numpy array: The rotated atomic positions with shape (... , 3)

    """

    # handle theta
    if theta is None:
        theta = np.random.uniform(-np.pi / 2., np.pi / 2., size=3)

    # rotation matrix
    R = euler_rotation_matrix(theta)

    return np.dot(atoms, R)


def padaxis(array, new_size, axis, pad_value=0, pad_right=True):
    """
    Padds one axis of an array to a new size
    This is just a wrapper for np.pad, more usefull when only padding a single axis

    Parameters
    ----------
    array: ndarray
        the array to pad

    new_size: int
        the new size of the specified axis

    axis: int
        axis along which to pad

    pad_value: float or int, optional(default=0)
        pad value

    pad_right: bool, optional(default=True)
        if True pad on the right side, otherwise pad on left side

    Returns
    -------
    ndarray: padded array

    """
    add_size = new_size - array.shape[axis]
    assert add_size >= 0, 'Cannot pad dimension {0} of size {1} to smaller size {2}'.format(
        axis, array.shape[axis], new_size)
    pad_width = [(0, 0)] * len(array.shape)

    #pad after if int is provided
    if pad_right:
        pad_width[axis] = (0, add_size)
    else:
        pad_width[axis] = (add_size, 0)

    return np.pad(array,
                  pad_width=pad_width,
                  mode='constant',
                  constant_values=pad_value)

def standardize_batch(batch_content, return_mask=False):
    '''
    Helper function that pads all sequential data in a batch to be the same length

    Parameters
    ----------
    batch_content: list
        a list of contents, where each element of the list is a sequential data

    return_mask: bool
        whether or not return the mask for valid data/padding
    '''
    if type(batch_content[0]) is np.ndarray:
        all_len = [len(item) for item in batch_content]
        maximal_len = np.max(all_len)
        batch_content = np.array([padaxis(arr, maximal_len, 0) for arr in batch_content])
    else:
        assert not return_mask
        batch_content = np.array(batch_content)
    if return_mask:
        mask = np.zeros((len(batch_content), maximal_len))
        for idx, length in enumerate(all_len):
            mask[idx, :length] = 1
        return batch_content, mask
    else:
        return batch_content

def check_data_consistency(data):
    """Check that various arrays in data dictionary are consistent.

    Parameters
    ----------
    data : dict
        The dictionary containing 'R', 'Z', 'N', 'E', 'F', and 'RXN' keys and corresponding
        array values.

    """
    # check data keys
    keys = sorted(['R', 'Z', 'N', 'E', 'F', 'RXN'])
    if sorted(data.keys()) != keys and sorted(data.keys()) != sorted(['CE', 'CF', 'E', 'F', 'N', 'R', 'RXN', 'Z']):
        raise ValueError(f"Data keys {sorted(data.keys())} does not match expected {keys} key!")

    # check data arrays shape
    if data['Z'].ndim != 2:
        raise ValueError("The data['Z'] ndim {data['Z'].dim} != 2")
    n_data, n_atom = data['Z'].shape

    if data['R'].shape != (n_data, n_atom, 3):
        raise ValueError(f"The data['R'] shape {data['R'].shape} != {(n_data, n_atom, 3)}")

    if data['E'].shape != (n_data, 1):
        raise ValueError(f"The data['E'] shape {data['E'].shape} != {(n_data, 1)}")

    if data['F'].shape != (n_data, n_atom, 3):
        raise ValueError(f"The data['F'] shape {data['F'].shape} != {(n_data, n_atom, 3)}")

    if data['N'].shape != (n_data, 1):
        raise ValueError(f"The data['N'] shape {data['N'].shape} != {(n_data, 1)}")

    if data['RXN'].shape != (n_data, 1):
        raise ValueError(f"The data['RXN'] shape {data['RXN'].shape} != {(n_data, 1)}")


def combine_rxn_arrays(rxns, n_max=6):
    """Combine reaction data dictionaries into one dictionary.

    Parameters
    ----------
    rxts : list
        List of dictionaries containing various datasets.
    n_max : int
        Maximum number of atoms in reactions.

    """
    # data dictionary to store combined data
    data = {'R': None, 'Z': None, 'N': None, 'E': None, 'F': None, 'RXN': None}
    n_total = 0

    for index, rxn in enumerate(rxns):
        # check reaction keys
        if sorted(rxn.keys()) != sorted(data.keys()):
            raise ValueError(f"RXN {index} keys={sorted(rxn.keys())} != {sorted(data.keys())}")

        # check consistency of data arrays shape
        check_data_consistency(rxn)

        # get number of data points and number of atoms
        n_data, n_atom = rxn['Z'].shape

        # check n_atom
        if n_atom > n_max:
            raise ValueError(f"RXN {index} has {n_atom} atoms which is greater than n_max={n_max}")

        for key, value in rxn.items():
            # pad arrays which depend on the number of atoms
            if key in ['R', 'Z', 'F'] and n_atom < n_max:
                value = padaxis(value, n_max, 1, pad_value=0, pad_right=True)
            # concatenate data
            if index == 0:
                data[key] = np.copy(value)
            else:
                data[key] = np.concatenate([data[key], value], axis=0)
        n_total += n_data

    # check consistency of data arrays shape
    check_data_consistency(data)

    return data


def write_data_npz(data, fname):
    """Save arrays into a single compressed file using the keywords given.

    Parameters
    ----------
    data : dict
        The dictionary containing 'R', 'Z', 'N', 'E', 'F', and 'RXN' keys and array values.
    fname : str
        Filename where the data will be saved.

    """
    if not isinstance(data, dict):
        raise ValueError('Argument data is supposed to be a dictionary!')
    # check consistency of data arrays shape
    check_data_consistency(data)
    # save data
    print(f"Generate {fname} containing {data['Z'].shape[0]} data points\n")
    np.savez_compressed(fname, **data)


def get_data_subset(data, rxn_num):
    """Return data dictionary corresponding to only specified reaction number.

    Parameters
    ----------
    data : dict
        The dictionary containing 'R', 'Z', 'N', 'E', 'F', and 'RXN' keys and array values.
    rxn_num : str
        The reaction number specified as a string of length two, appended with zero, if needed.

    """
    if not isinstance(rxn_num, str) or len(rxn_num) != 2:
        raise ValueError("Argument rxn_num is expected to be a string of length 2!")

    # check consistency of data arrays shape
    check_data_consistency(data)

    # check whether rxn exists in data
    rxn_nums = np.unique(data['RXN'])
    if rxn_num not in rxn_nums:
        raise ValueError(f"Reaction {rxn_num} does not exist among existing {rxn_nums} rxns!")
    if len(rxn_nums) == 1 and rxn_nums[0] == rxn_num:
        return data

    mask = np.where(data['RXN'] == rxn_num)[0]
    sub_data = {}
    for key, value in data.items():
        if key in ['Z', 'N', 'RXN', 'E']:
            value = value[mask, :]
        elif key in ['R', 'F']:
            value = value[mask, :, :]
        sub_data[key] = value

    # check consistency of sub data arrays
    check_data_consistency(sub_data)

    return sub_data


def get_data_remove_appended_zero(data):
    """Remove appended zeros to array values of data dictionary.

    Parameters
    ----------
    data : dict
        The dictionary containing 'R', 'Z', 'N', 'E', 'F', and 'RXN' keys and array values.
        If data contains information on more than one reaction, an error is raised.

    """

    # check consistency of data arrays shape
    check_data_consistency(data)

    # get reactions numbers included in data
    rxn_nums = np.unique(data['RXN'])
    print(f"Reactions contained in data: {rxn_nums}")
    if len(rxn_nums) != 1:
        if not len(rxn_nums) ==2 and rxn_nums[0][:-1] == rxn_nums[1][:-1]: #2 spin state reaction
            raise ValueError(f"Cannot remove appended zero as {rxn_nums} rxns were found!")

    # identify unique number of atoms
    n_atom = np.unique(data['N'])
    if len(n_atom) != 1:
        raise ValueError(f"Cannot remove appended zero as {n_atom} number of atoms were found!")
    n_atom = n_atom[0]

    # identify atomic numbers
    nums = np.unique(data['Z'], axis=0)
    if len(nums) != 1:
        raise ValueError(f"Cannot remove appended zero as {nums} atomic numbers were found!")
    if np.count_nonzero(nums[0] != 0) != n_atom:
        raise ValueError(f"Cannot remove appended zero as number of non-zero atomic numbers does "
                         "not match number of atoms!")

    new_data = {}
    for key, value in data.items():
        if key == 'Z':
            value = value[:, :n_atom]
        elif key in ['R', 'F']:
            value = value[:, :n_atom, :]
        new_data[key] = value

    # check consistency of new data arrays
    check_data_consistency(new_data)

    return new_data



import torch
from torch import nn


class ScaleShift(nn.Module):
    r"""Scale and shift layer for standardization.

    Credit : https://github.com/atomistic-machine-learning/schnetpack/blob/master/src/schnetpack/nn/base.py under the MIT License.

    .. math::
       y = x \times \sigma + \mu

    Parameters
    ----------
    mean: torch.Tensor
        mean value :math:`\mu`.

    stddev: torch.Tensor
        standard deviation value :math:`\sigma`.

    Copyright: https://github.com/atomistic-machine-learning/schnetpack/blob/master/src/schnetpack/nn/base.py
    """
    def __init__(self, mean, stddev):
        super(ScaleShift, self).__init__()
        self.register_buffer("mean", mean)
        self.register_buffer("stddev", stddev)

    def forward(self, input):
        """Compute layer output.

        Parameters
        ----------
        input: torch.Tensor
            input data.

        Returns
        -------
        torch.Tensor: layer output.

        """
        y = input * self.stddev + self.mean
        return y


class TrainableScaleShift(nn.Module):
    r"""Trainable scale and shift layer for standardization. Each atom type uses its dedicated scale and shift values

    .. math::
       y = x \times \sigma + \mu

    Parameters
    ----------
    max_z: the maximum atomic number for the molecule
    mean: torch.Tensor
        mean value :math:`\mu`.

    stddev: torch.Tensor
        standard deviation value :math:`\sigma`.

    """
    def __init__(self, max_z, initial_mean=None, initial_stddev=None):
        super(TrainableScaleShift, self).__init__()
        if initial_mean is not None:
            mean = nn.Parameter(torch.zeros(max_z) + initial_mean, requires_grad=True)
        else:
            mean = nn.Parameter(torch.zeros(max_z), requires_grad=True)
        if initial_stddev is not None:
            stddev = nn.Parameter(torch.ones(max_z) * initial_stddev, requires_grad=True)
        else:
            stddev = nn.Parameter(torch.ones(max_z), requires_grad=True)
        self.register_parameter('mean', mean)
        self.register_parameter('stddev', stddev)


    def forward(self, input_energies, z):
        """Compute layer output.

        Parameters
        ----------
        input_energies: torch.Tensor
            input atomic energies.

        z: torch.Tensor
            input atomic numbers

        Returns
        -------
        torch.Tensor: layer output.

        """
        selected_mean = self.mean[z][...,None]
        selected_stddev = self.stddev[z][...,None]
        y = input_energies * selected_stddev + selected_mean
        return y


"""
A compilation of modules that help to find closest neighbours of each atom in a molecule.
Each Molecule is represented as a dictionary with following keys:
    - atoms: atomic positions with shape (n_atom, 3)
    - z: atomic numbers with shape (n_atoms, 1)
    - cell: unit cell with shape (3,3)
    - atom_prop: atomic property with shape (n_atoms, n_atomic_prop)
    - mol_prop: molecular property with shape (1, n_mol_prop)

"""
import numpy as np
from ase import Atoms
from ase.neighborlist import neighbor_list

from newtonnet.utils import padaxis

class ExtensiveEnvironment(object):
    """
    Provide atomic environment of an array of atoms and their atomic numbers.
    No cutoff, No periodic boundary condition

    Parameters
    ----------
    max_n_neighbors: int, optional (default: None)
        maximum number of neighbors to pad arrays if they have less elements
        if None, it will be ignored (e.g., in case all atoms have same length)

    """
    def __init__(self, max_n_neighbors=None):
        if max_n_neighbors is None:
            max_n_neighbors = 0
        self.max_n_neighbors = max_n_neighbors

    def _check_shapes(self, Rshape, Zshape):
        if Rshape[0] != Zshape[0]:
            msg = "@ExtensiveEnvironment: atoms and atomic_numbers must have same dimension 0 (n_data)."
            raise ValueError(msg)

        if Rshape[2] != 3:
            msg = "@ExtensiveEnvironment: atoms must have 3 coordinates at dimension 2."
            raise ValueError(msg)

        if Rshape[1] != Zshape[1]:
            msg = "@ExtensiveEnvironment: atoms and atomic_numbers must have same dimension 1 (n_atoms)."
            raise ValueError(msg)

    def get_environment(self, positions, atomic_numbers):
        """
        This function finds atomic environments extensively for each atom in the MD snapshot.

        Parameters
        ----------
        positions: ndarray
            A 3D array of atomic positions in XYZ coordinates with shape (D, A, 3), where
            D is number of snapshots of data and A is number of atoms per data point

        atomic_numbers: ndarray
            A 2D array of atomic numbers with shape (D, A)

        Returns
        -------
        ndarray: 3D array of neighbors with shape (D, A, A-1)
        ndarray: 3D array of neighbor mask with shape (D, A, A-1)
        ndarray: 2D array of atomic mask for atomic energies (D, A)

        """
        n_data = positions.shape[0]  # D
        n_atoms = positions.shape[1]  # A

        self._check_shapes(positions.shape, atomic_numbers.shape)

        # 2d array of all indices for all atoms in a single data point
        N = np.tile(np.arange(n_atoms), (n_atoms, 1))  # (A, A)

        # remove the diagonal self indices
        neighbors = N[~np.eye(n_atoms, dtype=bool)].reshape(n_atoms,
                                                        -1)  # (A, A-1)
        neighbors = np.repeat(neighbors[np.newaxis, ...], n_data, axis=0)  # (D, A, A-1)

        # mask based on zero atomic_numbers
        mask = np.ones_like(atomic_numbers)                 #(D, A)
        mask[np.where(atomic_numbers == 0)] = 0
        max_atoms = np.sum(mask, axis=1)

        neighbor_mask = (neighbors < np.tile(max_atoms.reshape(-1,1), n_atoms-1)[:,None,:]).astype('int')
        neighbor_mask *= mask[:,:,None]  # (D,A,A-1)
        neighbors *= neighbor_mask  # (D,A,A-1)

        # atomic numbers
        # atomic numbers are already in correct shape

        if n_atoms < self.max_n_neighbors:
            neighbors = padaxis(neighbors,
                                self.max_n_neighbors,
                                axis=-1,
                                pad_value=-1)  # (D, A, N)
            atomic_numbers = padaxis(atomic_numbers,
                                     self.max_n_neighbors,
                                     axis=-1,
                                     pad_value=0)  # (D, A, N)

        return neighbors, neighbor_mask, mask, None, None


class PeriodicEnvironment(object):
    """
    Provide atomic environment of an array of atoms and their atomic numbers with cutoff and periodic boundary condition.


    Parameters
    ----------
    max_n_neighbors: int, optional (default: None)
        maximum number of neighbors to pad arrays if they have less elements
        if None, it will be ignored (e.g., in case all atoms have same length)

    cutoff: float
        the cutoff value to be used for finding neighbors
    """
    def __init__(self, max_n_neighbors=None, cutoff=7.0):
        if max_n_neighbors is None:
            max_n_neighbors = 0
        self.max_n_neighbors = max_n_neighbors
        self.cutoff = cutoff

    def _check_shapes(self, Rshape, Zshape):
        if Rshape[0] != Zshape[0]:
            msg = "@PeriodicEnvironment: atoms and atomic_numbers must have same dimension 0 (n_data)."
            raise ValueError(msg)

        if Rshape[2] != 3:
            msg = "@PeriodicEnvironment: atoms must have 3 coordinates at dimension 2."
            raise ValueError(msg)

        if Rshape[1] != Zshape[1]:
            msg = "@PeriodicEnvironment: atoms and atomic_numbers must have same dimension 1 (n_atoms)."
            raise ValueError(msg)

    def get_environment(self, positions, atomic_numbers, lattice):
        """
        This function finds atomic environments extensively for each atom in the MD snapshot using the ASE package.

        Parameters
        ----------
        positions: ndarray
            A 3D array of atomic positions in XYZ coordinates with shape (D, A, 3), where
            D is number of snapshots of data and A is number of atoms per data point

        atomic_numbers: ndarray
            A 2D array of atomic numbers with shape (D, A)

        lattice: ndarray
            A 2D array with shape (D, 9), where the second axis can be reshaped into (3x3) that 
            represents the 3 lattice vectors 

        Returns
        -------
        ndarray: 3D array of neighbors with shape (D, A, N), where N is either the maximum number of neighbors in 
            current batch of data, or the predefined max_n_neighbors
        ndarray: 3D array of neighbor mask with shape (D, A, N)
        ndarray: 2D array of atomic mask for atomic energies (D, A)
        ndarray: 3D array of neighbor distances with shape (D, A, N)
        ndarray: 4D array of neighbor atom arrays with shape (D, A, N, 3)
        """
        n_data = positions.shape[0]  # D
        n_atoms = positions.shape[1]  # A
        lattice = lattice.reshape((-1, 3, 3))

        self._check_shapes(positions.shape, atomic_numbers.shape)

        staggered_neighbors = [[[] for _ in range(n_atoms)] for _ in range(n_data)]
        staggered_distances = [[[] for _ in range(n_atoms)] for _ in range(n_data)]
        staggered_distance_vectors = [[[] for _ in range(n_atoms)] for _ in range(n_data)]
        neighbor_count = np.zeros((n_data, n_atoms), dtype=int)

        # mask based on zero atomic_numbers
        mask = np.ones_like(atomic_numbers)                 #(D, A)
        mask[np.where(atomic_numbers == 0)] = 0

        for idx_data in range(n_data):
            molecule_Rs = positions[idx_data, mask[idx_data] == 1]
            molecule_Zs = atomic_numbers[idx_data, mask[idx_data] == 1]
            molecule_lattice = lattice[idx_data]

            ase_molecule = Atoms(molecule_Zs, positions=molecule_Rs, cell=molecule_lattice, pbc=True)
            for i, j, dist, vec in zip(*neighbor_list("ijdD", ase_molecule, self.cutoff)):
                staggered_neighbors[idx_data][i].append(j)
                staggered_distances[idx_data][i].append(dist)
                staggered_distance_vectors[idx_data][i].append(vec)
                neighbor_count[idx_data][i] += 1

        max_N = np.max(neighbor_count.max(), self.max_n_neighbors)
        neighbors = np.zeros((n_data, n_atoms, max_N))
        distances = np.zeros((n_data, n_atoms, max_N))
        distance_vectors = np.zeros((n_data, n_atoms, max_N, 3))
        neighbor_mask = np.zeros((n_data, n_atoms, max_N))

        for i in range(n_data):
            for j in range(n_atoms):
                if neighbor_count[i, j] > 0:
                    neighbors[i, j, :neighbor_count[i, j]] = staggered_neighbors[i][j]
                    distances[i, j, :neighbor_count[i, j]] = staggered_distances[i][j]
                    distance_vectors[i, j, :neighbor_count[i, j]] = staggered_distance_vectors[i][j]
                    neighbor_mask[i, j, :neighbor_count[i, j]] = 1


        # # 2d array of all indices for all atoms in a single data point
        # N = np.tile(np.arange(n_atoms), (n_atoms, 1))  # (A, A)

        # # remove the diagonal self indices
        # neighbors = N[~np.eye(n_atoms, dtype=bool)].reshape(n_atoms,
        #                                                 -1)  # (A, A-1)
        # neighbors = np.repeat(neighbors[np.newaxis, ...], n_data, axis=0)  # (D, A, A-1)

        # max_atoms = np.sum(mask, axis=1)

        # neighbor_mask = (neighbors < np.tile(max_atoms.reshape(-1,1), n_atoms-1)[:,None,:]).astype('int')
        # neighbor_mask *= mask[:,:,None]  # (D,A,A-1)
        # neighbors *= neighbor_mask  # (D,A,A-1)

        # # atomic numbers
        # # atomic numbers are already in correct shape

        # if n_atoms < self.max_n_neighbors:
        #     neighbors = padaxis(neighbors,
        #                         self.max_n_neighbors,
        #                         axis=-1,
        #                         pad_value=-1)  # (D, A, N)
        #     atomic_numbers = padaxis(atomic_numbers,
        #                              self.max_n_neighbors,
        #                              axis=-1,
        #                              pad_value=0)  # (D, A, N)

        return neighbors, neighbor_mask, mask, distances, distance_vectors



import torch
from torch import nn
from torch.autograd import grad

from newtonnet.layers import Dense
from newtonnet.layers.shells import ShellProvider
from newtonnet.layers.scalers import ScaleShift, TrainableScaleShift
from newtonnet.layers.cutoff import CosineCutoff, PolynomialCutoff
from newtonnet.layers.representations import RadialBesselLayer


class NewtonNet(nn.Module):
    """
    Molecular Newtonian Message Passing

    Parameters
    ----------
    resolution: int
        number of radial functions to describe interatomic distances

    n_features: int
        number of neurons in the latent layer. This number will remain fixed in the entire network except
        for the last fully connected network that predicts atomic energies.

    activation: function
        activation function from newtonnet.layers.activations
        you can aslo get it by string from newtonnet.layers.activations.get_activation_by_string

    n_interactions: int, default: 3
        number of interaction blocks
    dropout
    max_z
    cutoff
    cutoff_network
    normalizer
    normalize_atomic
    requires_dr
    device
    create_graph
    shared_interactions
    return_latent
    """

    def __init__(self,
                 resolution,
                 n_features,
                 activation,
                 n_interactions=3,
                 dropout=0.0,
                 max_z=10,
                 cutoff=5.0,
                 cutoff_network='poly',
                 normalizer=(0.0, 1.0),
                 normalize_atomic=False,
                 requires_dr=False,
                 device=None,
                 create_graph=False,
                 shared_interactions=False,
                 return_hessian=False,
                 layer_norm=False,
                 atomic_properties_only=False,
                 double_update_latent=True,
                 pbc=False,
                 aggregration='sum'):

        super(NewtonNet, self).__init__()

        self.requires_dr = requires_dr
        self.create_graph = create_graph
        self.normalize_atomic = normalize_atomic
        self.return_hessian = return_hessian
        self.pbc = pbc

        # test

        shell_cutoff = None
        if pbc:
            # make the cutoff here a little bit larger so that it can be handled with differentiable cutoff layer in interaction block
            shell_cutoff = cutoff * 1.1

        self.shell = ShellProvider(return_vecs=True, normalize_vecs=True, pbc=pbc, cutoff=shell_cutoff)
        self.distance_expansion = RadialBesselLayer(
            resolution, cutoff, device=device
        )

        # atomic embedding
        self.n_features = n_features
        self.embedding = nn.Embedding(max_z, n_features, padding_idx=0)

        # d1 message
        self.n_interactions = n_interactions
        if shared_interactions:
            # use the same message instance (hence the same weights)
            self.dycalc = nn.ModuleList(
                [
                    DynamicsCalculator(
                        n_features=n_features,
                        resolution=resolution,
                        activation=activation,
                        cutoff=cutoff,
                        cutoff_network=cutoff_network,
                        double_update_latent=double_update_latent
                    )
                ]
                * n_interactions
            )
        else:
            # use one SchNetInteraction instance for each interaction
            self.dycalc = nn.ModuleList(
                [
                    DynamicsCalculator(
                        n_features=n_features,
                        resolution=resolution,
                        activation=activation,
                        cutoff=cutoff,
                        cutoff_network=cutoff_network,
                        double_update_latent=double_update_latent
                    )
                    for _ in range(n_interactions)
                ]
            )

        # layer norm
        self.layer_norm = layer_norm
        if layer_norm:
            self.norm = nn.ModuleList([nn.LayerNorm(n_features) for _ in range(n_interactions)])

        # final dense network
        self.atomic_energy = AtomicEnergy(n_features, activation, dropout)

        self.normalize_atomic = normalize_atomic
        if normalize_atomic:
            self.inverse_normalize = TrainableScaleShift(max_z)
        else:
            if type(normalizer) is dict:
                self.inverse_normalize = nn.ModuleDict(
                    {str(atom_num): ScaleShift(
                        mean=torch.tensor(normalizer[atom_num][0],
                                          device=device),
                        stddev=torch.tensor(normalizer[atom_num][1],
                                            device=device)) for atom_num in normalizer})
            else:
                self.inverse_normalize = ScaleShift(
                    mean=torch.tensor(normalizer[0],
                                      device=device),
                    stddev=torch.tensor(normalizer[1],
                                        device=device))

        self.atomic_properties_only = atomic_properties_only
        self.aggregration = aggregration

    def forward(self, data):

        Z = data['Z']
        R = data['R']
        N = data['N']
        NM = data['NM']
        AM = data['AM']
        if "lattice" in data:
            lattice = data['lattice']
        else:
            lattice = None

        # initiate main containers
        a = self.embedding(Z)  # B,A,nf
        f_dir = torch.zeros_like(R)  # B,A,3
        f_dynamics = torch.zeros(R.size() + (self.n_features,), device=R.device)  # B,A,3,nf
        r_dynamics = torch.zeros(R.size() + (self.n_features,), device=R.device)  # B,A,3,nf
        e_dynamics = torch.zeros_like(a)  # B,A,nf

        # require grad
        if self.requires_dr:
            R.requires_grad_()

        # compute distances (B,A,N) and distance vectors (B,A,N,3)
        if 'D' in data:
            distances = data['D']
            distance_vector = data['V']
        else:
            distances, distance_vector, N, NM = self.shell(R, N, NM, lattice)

        # comput d1 representation (B, A, N, G)
        rbf = self.distance_expansion(distances)

        # compute interaction block and update atomic embeddings
        for i_interax in range(self.n_interactions):
            # print('iter: ', i_interax)

            # messages
            a, f_dir, f_dynamics, r_dynamics, e_dynamics = self.dycalc[i_interax](a, rbf, distances, distance_vector, N,
                                                                                  NM,
                                                                                  f_dir, f_dynamics, r_dynamics,
                                                                                  e_dynamics
                                                                                  )  # B,A,f  # B,A,N,f

            if self.layer_norm:
                a = self.norm[i_interax](a)

        # When using the network to obtain atomic properties only
        if self.atomic_properties_only:
            Ai = self.atomic_energy(a)
            if self.normalize_atomic:
                Ai = self.inverse_normalize(Ai, Z)
            else:
                for atomic_type in self.inverse_normalize:
                    atomic_filter = Z == int(atomic_type)
                    Ai[atomic_filter] = self.inverse_normalize[atomic_type](Ai[atomic_filter])
            return {'Ai': Ai}

        # output net
        Ei = self.atomic_energy(a)
        if self.normalize_atomic:
            Ei = self.inverse_normalize(Ei, Z)

        # inverse normalize
        Ei = Ei * AM[..., None]  # (B,A,1)
        if self.aggregration == 'sum':
            E = torch.sum(Ei, 1)  # (B,1)
        elif self.aggregration == 'mean':
            E = torch.mean(Ei, 1)
        elif self.aggregration == 'max':
            E = torch.max(Ei, 1).values
        if not self.normalize_atomic:
            E = self.inverse_normalize(E)

        # if self.return_hessian:
        #     return E

        if self.requires_dr:
            if self.return_hessian:
                dE = grad(E, R, grad_outputs=torch.ones(E.shape[0], 1, device=R.device), create_graph=True, retain_graph=True)[0]
                ddE = torch.zeros(E.shape[0], R.shape[1], R.shape[2], R.shape[1], R.shape[2], device=R.device)
                for A_ in range(R.shape[1]):
                    for X_ in range(3):
                        ddE[:, A_, X_, :, :] = grad(dE[:, A_, X_], R, grad_outputs=torch.ones(E.shape[0], device=R.device), create_graph=False, retain_graph=True)[0]
                # ddE = torch.stack([grad(dE, R, grad_outputs=V, create_graph=True, retain_graph=True, allow_unused=True)[0] for V in torch.eye(R.shape[1] * R.shape[2], device=R.device).reshape((-1, 1, R.shape[1], R.shape[2])).repeat(1, R.shape[0], 1, 1)])
                # ddE = torch.vmap(lambda V: grad(dE, R, grad_outputs=V, create_graph=True, retain_graph=True))(torch.eye(R.shape[1] * R.shape[2], device=R.device).reshape((-1, 1, R.shape[1], R.shape[2])).repeat(1, R.shape[0], 1, 1))
                # ddE = ddE.permute(1,2,3,0).unflatten(dim=3, sizes=(-1, 3))
            else:
                dE = grad(E, R, grad_outputs=torch.ones(E.shape[0], 1, device=R.device), create_graph=self.create_graph, retain_graph=True)[0]
            dE = -1.0 * dE
        else:
            dE = data['F']

        if self.return_hessian:
            return {'R': R, 'E': E, 'F': dE, 'H': ddE, 'Ei': Ei, 'F_latent': f_dir}
        else:
            return {'R': R, 'E': E, 'F': dE, 'Ei': Ei, 'F_latent': f_dir}


class DynamicsCalculator(nn.Module):

    def __init__(
            self,
            n_features,
            resolution,
            activation,
            cutoff,
            cutoff_network,
            double_update_latent=True,
            epsilon=1e-8
    ):
        super(DynamicsCalculator, self).__init__()

        self.n_features = n_features
        self.epsilon = epsilon

        # non-directional message passing
        self.phi_rbf = Dense(resolution, n_features, activation=None)

        self.phi_a = nn.Sequential(
            Dense(n_features, n_features, activation=activation),
            Dense(n_features, n_features, activation=None),
        )

        # cutoff layer used in interaction block
        if cutoff_network == 'poly':
            self.cutoff_network = PolynomialCutoff(cutoff, p=9)
        elif cutoff_network == 'cosine':
            self.cutoff_network = CosineCutoff(cutoff)

        # directional message passing
        self.phi_f = Dense(n_features, 1, activation=None, bias=False)
        self.phi_f_scale = nn.Sequential(
            Dense(n_features, n_features, activation=activation),
            Dense(n_features, n_features, activation=None),
        )
        self.phi_r = nn.Sequential(
            Dense(n_features, n_features, activation=None),
        )
        self.phi_r = nn.Sequential(
            Dense(n_features, n_features, activation=activation, xavier_init_gain=0.001),
            Dense(n_features, n_features, activation=None),
        )
        self.phi_r_ext = nn.Sequential(
            Dense(n_features, n_features, activation=activation, bias=False),
            Dense(n_features, n_features, activation=None, bias=False),
        )

        self.phi_e = nn.Sequential(
            Dense(n_features, n_features, activation=activation),
            Dense(n_features, n_features, activation=None)
        )

        self.double_update_latent = double_update_latent

    def gather_neighbors(self, inputs, N):
        n_features = inputs.size()[-1]
        n_dim = inputs.dim()
        b, a, n = N.size()  # batch, atoms, neighbors size

        if n_dim == 3:
            N = N.view(-1, a * n, 1)  # B,A*N,1
            N = N.expand(-1, -1, n_features)
            out = torch.gather(inputs, dim=1, index=N)
            return out.view(b, a, n, n_features)  # B,A,N,n_features

        elif n_dim == 4:
            N = N.view(-1, a * n, 1, 1)  # B,A*N,1,1
            N = N.expand(-1, -1, 3, n_features)
            out = torch.gather(inputs, dim=1, index=N)
            return out.view(b, a, n, 3, n_features)  # B,A,N,3,n_features

    def sum_neighbors(self, x, mask, dim=2, avg=False):
        """

        Parameters
        ----------
        x: torch.tensor
            usually of shape B,A,N,nf
        mask: torch.tensor
            usually of shape B,A,N
        dim: int
            the dimension to sum

        avg: bool
            if True, returns the average output by dividing the sum by number of neighbors.

        Returns
        -------

        """
        dim_diff = x.dim() - mask.dim()
        for _ in range(dim_diff):
            mask = mask.unsqueeze(-1)

        x = x * mask
        out = torch.sum(x, dim=dim)

        if avg:
            n_atoms = torch.sum(mask, dim)
            n_atoms = torch.max(n_atoms, other=torch.ones_like(n_atoms))
            out = out / n_atoms

        return out

    def forward(self, a, rbf, distances, distance_vector, N, NM,
                f_dir, f_dynamics, r_dynamics, e_dynamics
                ):

        # map decomposed distances
        rbf_msij = self.phi_rbf(rbf)  # B,A,N,nf

        # cutoff
        C = self.cutoff_network(distances)
        rbf_msij = rbf_msij * C.unsqueeze(-1)

        # map atomic features
        a_msij = self.phi_a(a)  # B,A,3*nf

        # copy central atom features for the element-wise multiplication
        ai_msij = a_msij.repeat(1, 1, rbf_msij.size(2))
        ai_msij = ai_msij.view(rbf_msij.size())  # B,A,N,nf

        # look up neighboring atoms features based on the schnet contiuous filter implementation
        aj_msij = self.gather_neighbors(a_msij, N)  # B,A,N,nf

        # symmetric feature multiplication
        mij = rbf_msij * aj_msij
        msij = mij * ai_msij

        # update a with invariance
        if self.double_update_latent:
            a = a + self.sum_neighbors(msij, NM, dim=2)

        # Dynamics: Forces
        # print('msij:', msij.shape, msij[0,0])
        F_ij = self.phi_f(msij) * distance_vector  # B,A,N,3
        F_i_dir = self.sum_neighbors(F_ij, NM, dim=2)  # B,A,3
        f_dir = f_dir + F_i_dir

        F_ij = self.phi_f_scale(msij).unsqueeze(-2) * F_ij.unsqueeze(-1)  # B,A,N,3,nf
        # print('F_ij:', F_ij.shape, F_ij[0,0])
        F_i = self.sum_neighbors(F_ij, NM, dim=2)  # B,A,3,nf

        # dr
        dr_i = self.phi_r(a).unsqueeze(-2) * F_i  # B,A,3,nf

        dr_j = self.gather_neighbors(r_dynamics, N)  # B,A,N,3,nf
        dr_j = self.phi_r_ext(msij).unsqueeze(-2) * dr_j  # B,A,N,3,nf
        # print('dr_j:', dr_j.shape, dr_j[0,0])
        dr_ext = self.sum_neighbors(dr_j, NM, dim=2, avg=False)  # B,A,3,nf

        # update
        f_dynamics = f_dynamics + F_i
        r_dynamics = r_dynamics + dr_i + dr_ext

        # update energy
        de_i = -1.0 * torch.sum(f_dynamics * r_dynamics, dim=-2)  # B,A,nf
        de_i = self.phi_e(a) * de_i
        a = a + de_i
        e_dynamics = e_dynamics + de_i

        return a, f_dir, f_dynamics, r_dynamics, e_dynamics


class AtomicEnergy(nn.Module):

    def __init__(self, n_features, activation, dropout):
        super(AtomicEnergy, self).__init__()
        self.environment = nn.Sequential(
            Dense(n_features, 128,
                  activation=activation,
                  dropout=dropout,
                  norm=False),
            Dense(128, 64,
                  activation=activation,
                  dropout=dropout,
                  norm=False),
            Dense(64, 1,
                  activation=None,
                  dropout=0.0,
                  norm=False),
        )

    def forward(self, a):
        # update atomic features
        out = self.environment(a)

        return out



import os
import torch
import torch.nn.functional as F

import numpy as np
import plotly
import matplotlib.pyplot as plt
import plotly.express as px


class VisHook(object):
    """
    Base class for visualization hooks.
    """

    def create_grid(self, ndim, length, size):
        half_length = length / 2.
        offsets = torch.linspace(-half_length, half_length, size)
        if ndim == 3:
            grids = torch.meshgrid([offsets, offsets, offsets])
        elif ndim == 2:
            grids = torch.meshgrid([offsets, offsets])

        grid = torch.stack(grids, dim=-1)  # shape: G,G,G,3

        grid = grid.data.cpu().numpy()
        # grid = grid.reshape(-1, 2)
        return grid

    def set_output(self, show=True, path2save=None):
        """

        Parameters
        ----------
        show: bool
            If True, show the plot in the console.
        path: str
            path to the folder to save the figure.

        Returns
        -------

        """
        self.show = show
        self.path2save = path2save
        if path2save is not None:
            if not os.path.exists(path2save):
                os.makedirs(path2save)


    def get_2d_rot_mat(self, theta):
        """
        Affine transformation matrix for counter clockwise rotation around origin.

        Parameters
        ----------
        theta: float
            degree in float


        Returns
        -------

        """
        theta = np.radians(theta)
        theta = torch.tensor(theta)
        return torch.tensor([[torch.cos(theta), -torch.sin(theta), 0],
                             [torch.sin(theta), torch.cos(theta), 0]])

    def rot_grid(self, x, theta, dtype, device):
        rot_mat = self.get_2d_rot_mat(theta)
        rot_mat = rot_mat[None, ...].type(dtype).repeat(x.shape[0],1,1)
        grid = F.affine_grid(rot_mat, x.size()).type(dtype)
        grid = grid.to(device)
        x = F.grid_sample(x, grid)
        return x


class Vis3DVoxelCartesian(VisHook):
    def __init__(self, grid_size, grid_length):
        self.grid = self.create_grid(3, grid_length, grid_size)

    def run(self, input):
        assert input.shape == self.grid.shape


class VizMolVectors3D(VisHook):

    def __init__(self):
        pass

    def run(self, R, Z, F, dF, dR):
        """
        Parameters
        ----------
        R: ndarray
            shape: A,3
        Z: ndarray
            shape: A,
        F: ndarray
            shape: A,3
        dF: ndarray
            shape: A,3
        dR: ndarray
            shape: A,3

        """
        if isinstance(R, torch.Tensor):
            R = R.data.cpu().numpy()
            Z = Z.data.cpu().numpy()
            F = F.data.cpu().numpy()
            dF= dF.data.cpu().numpy()
            dR= dR.data.cpu().numpy()

        # move R
        R_mean = np.mean(R, axis=0)
        R = R - R_mean
        # xlim, ylim, zlim = np.max(R, axis=0)

        # normalize vectors
        fmax = np.linalg.norm(F, axis=1).max()
        fmax += 1e-8
        F = F / fmax
        # dF = dF / fmax
        # dR = dR / fmax

        F *= 2
        dF *= 4

        # fig = plt.figure()
        ax = plt.figure().add_subplot(projection='3d')

        # cpk coloring
        atom_color = {1: 'gray', 8: 'r', 6: 'k', 7: 'b', 16: 'y'}

        ax.quiver(R[:, 0], R[:, 1], R[:, 2], F[:, 0], F[:, 1], F[:, 2], color='g', **{'linewidth':1.5})  # , scale=21
        ax.quiver(R[:, 0], R[:, 1], R[:, 2], dF[:, 0], dF[:, 1], dF[:, 2], color='b',**{'linewidth':1.5})  # , scale=21
        if dR is not None and dR.ndim == dF.ndim:
            ax.quiver(R[:, 0], R[:, 1], R[:, 2], dR[:, 0], dR[:, 1], dR[:, 2], color='r')  # , scale=21
        ax.scatter(R[:, 0], R[:, 1], R[:, 2], color=[atom_color[z] for z in Z], s=250, alpha=0.8)

        ax.set_xlabel('X')
        ax.set_ylabel('Y')
        ax.set_zlabel('Z')

        # Hide grid lines
        ax.grid(False)
        # Hide axes ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])

        # plt.xlim(-5, 5)
        # plt.ylim(-5, 5)
        # ax.set_zlim(-5, 5)

        if self.show:
            plt.show()

        if self.path2save is not None:
            plt.savefig(os.path.join(self.path2save, "mol_3d_vectors.png"), dpi=300)
            plt.savefig(os.path.join(self.path2save, "mol_3d_vectors.eps"), dpi=300)






import os
import numpy as np
import pandas as pd
import torch
import time
import shutil
import matplotlib.pyplot as plt
from tqdm import tqdm

from torch.optim.lr_scheduler import ReduceLROnPlateau, LambdaLR
from torch import nn
from newtonnet.utils.utility import standardize_batch
from itertools import chain


class Trainer:
    """
    Parameters
    ----------
    """
    def __init__(self,
                 model,
                 loss_fn,
                 optimizer,
                 requires_dr,
                 device,
                 yml_path,
                 output_path,
                 script_name,
                 lr_scheduler,
                 energy_loss_w,
                 force_loss_w,
                 loss_wf_decay,
                 lambda_l1,
                 checkpoint_log=1,
                 checkpoint_val=1,
                 checkpoint_test=20,
                 checkpoint_model=1,
                 verbose=False,
                 training=True,
                 hooks=None,
                 mode="energy/force",
                 target_name=None,
                 force_latent=False):
        self.model = model
        self.loss_fn = loss_fn
        self.optimizer = optimizer
        self.requires_dr = requires_dr
        self.device = device
        self.energy_loss_w = energy_loss_w
        self.force_loss_w = force_loss_w
        self.wf_lambda = lambda epoch: np.exp(-epoch * loss_wf_decay)
        self.lambda_l1 = lambda_l1

        if type(device) is list and len(device) > 1:
            self.multi_gpu = True
        else:
            self.multi_gpu = False

        # outputs
        self._subdirs(yml_path, output_path, script_name)
        if training:

            # hooks
            if hooks:
                self.hooks = None
                self._hooks(hooks)

            # learning rate scheduler
            self._handle_scheduler(lr_scheduler, optimizer)
            self.lr_scheduler = lr_scheduler

        # checkpoints
        self.check_log = checkpoint_log
        self.check_val = checkpoint_val
        self.check_test = checkpoint_test
        self.check_model = checkpoint_model
        self.verbose = verbose

        # checkpoints
        self.epoch = 0  # number of epochs of any steps that model has gone through so far
        self.log_loss = {
            'epoch': [],
            'loss(MSE)': [],
            'lr': [],
            'time': []
        }
        if mode in ["energy/force", "energy"]:
            self.log_loss.update({
                'tr_E(MAE)': [],
                'tr_F(MAE)': [],
                'val_E(MAE)': [],
                'val_F(MAE)': [],
                'irc_E(MAE)': [],
                'irc_F(MAE)': [],
                'test_E(MAE)': [],
                'test_F(MAE)': []
            })
        elif mode == "atomic_properties":
            self.log_loss.update({
                "tr_err(RMSE)": [],
                "val_err(RMSE)": [],
                "test_err(RMSE)": []
            })
        self.best_val_loss = float("inf")
        self.mode = mode
        self.target_name = target_name
        self.force_latent = force_latent

    def _handle_scheduler(self, lr_scheduler, optimizer):

        if lr_scheduler[0] == 'plateau':
            self.scheduler = ReduceLROnPlateau(optimizer=optimizer,
                                               mode='min',
                                               patience=lr_scheduler[2],
                                               factor=lr_scheduler[3],
                                               min_lr=lr_scheduler[4])
        elif lr_scheduler[0] == 'decay':
            lambda1 = lambda epoch: np.exp(-epoch * lr_scheduler[1])
            self.scheduler = LambdaLR(optimizer=optimizer, lr_lambda=lambda1)

        else:
            raise NotImplemented('scheduler "%s" is not implemented yet.'%lr_scheduler[0])

    def _subdirs(self, yml_path, output_path, script_name):

        # create output directory and subdirectories
        path_iter = output_path[1]
        out_path = os.path.join(output_path[0], 'training_%i'%path_iter)
        while os.path.exists(out_path):
            path_iter+=1
            out_path = os.path.join(output_path[0],'training_%i'%path_iter)
        os.makedirs(out_path)
        self.output_path = out_path

        # self.val_out_path = os.path.join(self.output_path, 'validation')
        # os.makedirs(self.val_out_path)

        # subdir for computation graph
        self.graph_path = os.path.join(self.output_path, 'graph')
        if not os.path.exists(self.graph_path):
            os.makedirs(self.graph_path)

        # saved models
        self.model_path = os.path.join(self.output_path, 'models')
        if not os.path.exists(self.model_path):
            os.makedirs(self.model_path)

        script_out = os.path.join(self.output_path, 'run_scripts')
        os.makedirs(script_out)
        shutil.copyfile(yml_path, os.path.join(script_out,os.path.basename(yml_path)))
        shutil.copyfile(script_name, os.path.join(script_out,os.path.basename(script_name)))

    def _hooks(self, hooks):
        hooks_list = []
        if 'vismolvector3d' in hooks and hooks['vismolvector3d']:
            from combust.train.hooks import VizMolVectors3D

            vis = VizMolVectors3D()
            vis.set_output(True, None)
            hooks_list.append(vis)

        if len(hooks_list) > 0:
            self.hooks = hooks_list


    def print_layers(self):
        total_n_params = 0
        for name, param in self.model.named_parameters():
            if param.requires_grad:
                print(name, param.shape)
                if len(param.shape) > 1:
                    total_n_params += param.shape[0] * param.shape[1]
                else:
                    total_n_params += param.shape[0]
        print('\n total trainable parameters: %i\n' % total_n_params)

    def plot_grad_flow(self):
        ave_grads = []
        layers = []
        for n, p in self.model.named_parameters():
            if (p.requires_grad) and ("bias" not in n):
                # shorten names
                layer_name = n.split('.')
                layer_name = [l[:3] for l in layer_name]
                layer_name = '.'.join(layer_name[:-1])
                layers.append(layer_name)
                # print(layer_name, p.grad)
                if p.grad is not None:
                    ave_grads.append(p.grad.abs().mean().detach().cpu())
                else:
                    ave_grads.append(0)

        fig, ax = plt.subplots(1, 1)
        ax.plot(ave_grads, alpha=0.3, color="b")
        ax.hlines(0, 0, len(ave_grads) + 1, linewidth=1, color="k")
        plt.xticks(range(0, len(ave_grads), 1), layers, rotation="vertical")
        plt.xlim(xmin=0, xmax=len(ave_grads))
        plt.xlabel("Layers")
        plt.ylabel("average gradient")
        plt.title("Gradient flow: epoch#%i" %self.epoch)
        plt.grid(True)
        ax.set_axisbelow(True)

        file_name= os.path.join(self.graph_path,"avg_grad.png")
        plt.savefig(file_name, dpi=300,bbox_inches='tight')
        plt.close(fig)

    def store_checkpoint(self, input, steps):
        self.log_loss['epoch'].append(self.epoch)
        for k in input:
            self.log_loss[k].append(input[k])


        df = pd.DataFrame(self.log_loss)
        df.applymap('{:.5f}'.format).to_csv(os.path.join(
            self.output_path, 'log.csv'),
                                           index=False)

        print("[%d, %3d]" % (self.epoch, steps), end="")
        for k in input:
            print("%s: %.5f; " % (k, input[k]), end="")
        print("\n")
        #  loss_mse: %.5f; "
        #     "tr_E(MAE): %.5f; tr_F(MAE): %.5f; "
        #     "val_E(MAE): %.5f; val_F(MAE): %.5f; "
        #     "irc_E(MAE): %.5f; irc_F(MAE): %.5f; "
        #     "test_E(MAE): %.5f; test_F(MAE): %.5f; "
        #     "lr: %.9f; epoch_time: %.3f\n"
        #     % (self.epoch, steps, np.sqrt(input[0]),
        #        input[1], input[2], input[3], input[4], input[5], input[6],
        #        input[7], input[8], input[9], input[10]))

    def _optimizer_to_device(self):
        for state in self.optimizer.state.values():
            for k, v in state.items():
                if torch.is_tensor(v):
                    state[k] = v.to(self.device[0])

    def metric_se(self, preds, data):
        """
        squared error

        Parameters
        ----------
        preds: B,A,...
        data: B,A,...

        """
        diff = preds - data
        se = diff**2

        # diff_forces = preds[1] - batch_data.forces
        # diff_forces = diff_forces ** 2

        # err_sq = torch.sum(diff_energy) + torch.sum(diff_forces)

        return se

    def metric_ae(self, preds, data, divider=None):
        """absolute error"""
        if data.ndim < preds.ndim:
            data = data[:, None]
        ae = np.abs(preds - data)
        if divider is not None:
            if divider.ndim < ae.ndim:
                divider = divider[:, None]
            ae /= divider
        # diff_forces = preds[1] - batch_data.forces
        # diff_forces = diff_forces ** 2

        # err_sq = torch.sum(diff_energy) + torch.sum(diff_forces)

        return ae

    def metric_rmse(self, preds, data, mask=None):
        """Root mean square error"""
        se = np.square(preds.squeeze() - data)
        if mask is not None:
            se *= mask
            return np.sqrt(np.sum(se) / (np.sum(mask) + 1e-7))
        else:
            return np.sqrt(np.mean(se))

    def masked_average(self, y, atom_mask):
        """

        Parameters
        ----------
        y: numpy array
        atom_mask: numpy array

        Returns
        -------

        """
        # handle rotation-wise loader batch size mismatch
        if atom_mask.shape[0] > y.shape[0]:
            # assert atom_mask.shape[1] == y.shape[1]
            atom_mask = atom_mask.reshape(y.shape[0], -1, y.shape[1])  # B, n_rot, A
            atom_mask = atom_mask.mean(axis=1)

        # size = np.sum(atom_mask, axis=1, keepdims=True)
        # size = np.maximum(size, np.ones_like(size))
        # y = np.sum(y, axis=1)
        # y = y / size

        y = y[atom_mask!=0]

        return y

    def resume_model(self, path):
        checkpoint = torch.load(path)
        self.model.load_state_dict(checkpoint['model_state_dict'])
        self.optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        self.epoch = checkpoint['epoch']
        loss = checkpoint['loss']
        return loss

    def validation(self, name, generator, steps):

        self.model.eval()
        self.model.requires_dr = self.requires_dr

        val_error_energy = []
        val_error_force = []
        energy_pred = []
        force_pred = []
        e = []
        f = []
        ei = []
        fi = []
        AM = []
        RM = []  # rotation angles/matrix

        for val_step in range(steps):
            val_batch = next(generator)

            if self.hooks is not None and val_step == steps-1:
                self.model.return_intermediate = True
                val_preds = self.model(val_batch)

                hs = val_preds['hs']
                for iter in range(1,4):
                    self.hooks[0].run(val_batch['R'][0], val_batch['Z'][0],val_batch['F'][0],
                                   hs[iter][1][0],hs[iter][2][0])
                    R_ = val_batch['R'].data.cpu().numpy()
                    np.save(os.path.join(self.val_out_path, 'hs_%i_R_'%iter), R_)
                    Z_ = val_batch['Z'].data.cpu().numpy()
                    np.save(os.path.join(self.val_out_path, 'hs_%i_Z_'%iter), Z_)
                    if self.force_latent:
                        F_ = val_batch['F_latent'].data.cpu().numpy()
                    else:
                        F_ = val_batch['F'].data.cpu().numpy()
                    np.save(os.path.join(self.val_out_path, 'hs_%i_F_'%iter), F_)
                    dF_ = hs[iter][1].data.cpu().numpy()
                    np.save(os.path.join(self.val_out_path, 'hs_%i_dF_'%iter), dF_)
                    dR_ = hs[iter][2].data.cpu().numpy()
                    np.save(os.path.join(self.val_out_path, 'hs_%i_dR_'%iter), dR_)

                # self.model.return_intermediate = False
            else:
                val_preds = self.model(val_batch)

            if val_preds['E'].ndim == 3:
                E = val_batch["E"].unsqueeze(1).repeat(1,val_batch["Z"].shape[1],1)
            else:
                E = val_batch["E"]

            val_error_energy.append(self.metric_ae(
                val_preds['E'].detach().cpu().numpy(), E.detach().cpu().numpy(),
                divider=val_batch['NA'].detach().cpu().numpy() if 'NA' in val_batch else None))
            energy_pred.append(val_preds['E'].detach().cpu().numpy())
            e.append(E.detach().cpu().numpy())
            ei.append(val_preds['Ei'].detach().cpu().numpy())
            if self.mode == 'energy/force':
                if self.force_latent:
                    predicted_force = val_preds['F_latent'].detach().cpu().numpy()
                else:
                    predicted_force = val_preds['F'].detach().cpu().numpy()
                target_force = val_batch["F"].detach().cpu().numpy()
                val_error_force.append(self.metric_ae(
                predicted_force, target_force))

                force_pred.append(predicted_force)
                f.append(target_force)

            if 'dEi' in val_preds and val_preds['dEi'] is not None:
                fi.append(val_preds['dEi'].detach().cpu().numpy())
            AM.append(val_batch["AM"].detach().cpu().numpy())
            RM.append(val_batch["RM"].detach().cpu().numpy())


            if self.verbose:
                val_error_force_report = standardize_batch(list(chain(*val_error_force)))
                AM_report = standardize_batch(list(chain(*AM)))
                val_mae_force_report = np.mean(self.masked_average(val_error_force_report, AM_report))

                print(
                    "%s: %i/%i - E_loss(MAE): %.5f - F_loss(MAE): %.5f"
                    % (name, val_step, steps,
                       np.mean(np.concatenate(val_error_energy, axis=0)),
                       val_mae_force_report
                       ))

            del val_batch

        outputs = dict()
        AM = standardize_batch(list(chain(*AM)))
        outputs['AM'] = AM
        outputs['RM'] = np.concatenate(RM, axis=0)
        outputs['E_ae'] = np.concatenate(val_error_energy, axis=0)
        outputs['E_pred'] = np.concatenate(energy_pred,axis=0)
        outputs['E'] = np.concatenate(e, axis=0)
        outputs['Ei'] = standardize_batch(list(chain(*ei)))
        if len(fi) > 0:
            outputs['dEi'] = standardize_batch(list(chain(*fi)))
        if self.mode == 'energy/force':
            F_ae = standardize_batch(list(chain(*val_error_force)))
            outputs['F_ae_masked'] = self.masked_average(F_ae, AM)
            outputs['F_ae'] = F_ae
            outputs['F_pred'] = standardize_batch(list(chain(*force_pred)))
            outputs['F'] = standardize_batch(list(chain(*f)))
            outputs['total_ae'] = np.mean(outputs['E_ae']) + np.mean(outputs['F_ae_masked'])
        else:
            outputs['F_ae_masked'] = 0
            outputs['F_ae'] = 0
            outputs['F_pred'] = []
            outputs['F'] = 0
            outputs['total_ae'] = np.mean(outputs['E_ae'])


        return outputs

    def validation_atomic_properties(self, name, target_name, generator, steps):
        target_name = "Ai"
        if self.target_name is not None:
            target_name = self.target_name
        self.model.eval()
        self.model.requires_dr = False

        val_rmse = []
        pred = []
        R = []
        Z = []
        CS = []
        M = []
        labels = []
        # e = []
        # f = []
        # ei = []
        # fi = []
        AM = []
        RM = []  # rotation angles/matrix

        for val_step in range(steps):
            val_batch = next(generator)
            with torch.no_grad():
                val_preds = self.model(val_batch)[target_name]
            batch_target = val_batch[target_name]
            batch_mask = val_batch["M"]

            val_rmse.append(self.metric_rmse(
                val_preds.detach().cpu().numpy(),
                batch_target.detach().cpu().numpy(),
                batch_mask.detach().cpu().numpy()
            ))
            pred.append(val_preds.detach().cpu().numpy())
            R.append(val_batch["R"].detach().cpu().numpy())
            Z.append(val_batch["Z"].detach().cpu().numpy())
            CS.append(val_batch["CS"].detach().cpu().numpy())
            M.append(batch_mask.detach().cpu().numpy())
            if "labels" in val_batch:
                labels.extend(val_batch["labels"])

            AM.append(val_batch["AM"].detach().cpu().numpy())
            RM.append(val_batch["RM"].detach().cpu().numpy())


            if self.verbose:

                print(
                    "%s: %i/%i - %s_RMSE: %.5f"
                    % (name, val_step, steps,
                    target_name, np.mean(np.array(val_rmse))
                       ))

            del val_batch

        outputs = dict()
        AM = standardize_batch(list(chain(*AM)))
        outputs['AM'] = AM
        outputs['RM'] = np.concatenate(RM, axis=0)
        outputs['pred'] = standardize_batch(list(chain(*pred)))
        outputs['RMSE'] = np.mean(val_rmse)
        outputs["R"] = R
        outputs["Z"] = Z
        outputs["CS"] = CS
        outputs["M"] = M
        outputs["labels"] = labels
        return outputs

    def train(self,
              train_generator,
              epochs,
              steps,
              val_generator=None,
              val_steps=None,
              irc_generator=None,
              irc_steps=None,
              test_generator=None,
              test_steps=None,
              clip_grad=0):
        """
        The main function to train model for the given number of epochs (and steps per epochs).
        The implementation allows for resuming the training with different data and number of epochs.

        Parameters
        ----------
        epochs: int
            number of training epochs

        steps: int
            number of steps to call it an epoch (designed for nonstop data generators)


        """
        self.model.to(self.device[0])
        if self.multi_gpu:
            self.model = nn.DataParallel(self.model, device_ids=self.device)
        self._optimizer_to_device()

        running_val_loss = []
        last_test_epoch = 0
        for _ in tqdm(range(epochs)):
            t0 = time.time()

            # record total number of epochs so far
            self.epoch += 1

            # loss force weight decay
            w_f = self.force_loss_w * self.wf_lambda(self.epoch)

            # training
            running_loss = 0.0
            ae_energy = 0.0
            ae_force = 0.0
            rmse_ai = []
            n_data = 0
            n_atoms = 0
            self.model.train()
            self.model.requires_dr = self.requires_dr
            self.optimizer.zero_grad()
            # step_iterator = range(steps)
            # if not self.verbose:
            #     step_iterator = tqdm(step_iterator)

            for s in range(steps):
                self.optimizer.zero_grad()

                train_batch = next(train_generator)
                # self.model.module(train_batch)
                # preds = self.model.forward(train_batch)
                preds = self.model(train_batch)
                loss = self.loss_fn(preds, train_batch, self.model.parameters(),
                                    w_e=self.energy_loss_w,
                                    w_f=w_f,
                                    lambda_l1=self.lambda_l1)
                loss.backward()
                if clip_grad>0:
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), clip_grad)
                self.optimizer.step()
                # if (s+1)%4==0 or s==steps-1:
                #     self.optimizer.step()          # if in, comment out the one in the loop
                #     self.optimizer.zero_grad()     # if in, comment out the one in the loop

                current_loss = loss.detach().item()
                running_loss += current_loss

                atom_mask = train_batch["AM"].detach().cpu().numpy()
                n_atoms += np.sum(atom_mask)

                if self.mode in ["energy/force", "energy"]:
                    ae_energy += np.sum(self.metric_ae(
                        preds['E'].detach().cpu().numpy(),
                        train_batch["E"].detach().cpu().numpy(),
                        divider=train_batch['NA'].detach().cpu().numpy() if 'NA' in train_batch else None))

                    if self.mode == "energy/force":
                        ae_f = self.metric_ae(
                            preds['F'].detach().cpu().numpy(),
                            train_batch["F"].detach().cpu().numpy())
                        actual_ae_f = self.masked_average(ae_f, atom_mask)
                        ae_force += np.sum(actual_ae_f)
                    n_data += train_batch["E"].size()[0]
                    
                # n_atoms = train_batch.R.size()[1]

                    if self.verbose:
                        print(
                            "Train: Epoch %i/%i - %i/%i - loss: %.5f - running_loss(RMSE): %.5f - E(MAE): %.5f - F(MAE): %.5f"
                            % (self.epoch, epochs, s, steps, current_loss,
                            np.sqrt(running_loss / (s + 1)),
                            (ae_energy / (n_data)),
                            (ae_force / (n_atoms*3))
                            ))
                elif self.mode == "atomic_properties":
                    target_name = 'Ai'
                    if self.target_name is not None:
                        target_name = self.target_name
                    rmse_ai.append(np.mean(self.metric_rmse(
                        preds[target_name].detach().cpu().numpy(),
                        train_batch["CS"].detach().cpu().numpy(),
                        train_batch["M"].detach().cpu().numpy()
                    )))

                    n_data += train_batch["CS"].size()[0]

                    if self.verbose:
                        print(
                            "Train: Epoch %i/%i - %i/%i - loss: %.5f - running_loss(RMSE): %.5f - RMSE: %.5f"
                            % (self.epoch, epochs, s, steps, current_loss,
                            np.sqrt(running_loss / (s + 1)),
                            (np.mean(rmse_ai[-100:]))
                            ))
                del train_batch

            running_loss /= steps
            if self.mode in ["energy/force", "energy"]:
                ae_energy /= n_data
                ae_force /= (n_atoms * 3)
            elif self.mode == "atomic_properties":
                rmse_ai = np.mean(rmse_ai[-100:])

            # plots
            # self.plot_grad_flow()

            # validation
            val_error = float("inf")
            if self.mode in ["energy/force", "energy"]:
                val_mae_E = val_mae_F = 0
                if val_generator is not None and \
                    self.epoch % self.check_val == 0:

                    outputs = self.validation('valid', val_generator, val_steps)
                    if self.requires_dr:
                        val_error = self.energy_loss_w * np.mean(outputs['E_ae']) + \
                                    self.force_loss_w * np.mean(outputs['F_ae_masked'])
                    else:
                        val_error = self.energy_loss_w * np.mean(outputs['E_ae'])

                    val_mae_E = np.mean(outputs['E_ae'])
                    val_mae_F = np.mean(outputs['F_ae_masked'])
            elif self.mode == "atomic_properties":
                if val_generator is not None and \
                    self.epoch % self.check_val == 0:

                    outputs = self.validation_atomic_properties('valid', "CS", val_generator, val_steps)
                    val_error = outputs["RMSE"]

            # best model
            irc_mae_E = 0; irc_mae_F = 0
            test_mae_E = 0; test_mae_F = 0
            test_error = 0
            if self.best_val_loss > val_error:
                self.best_val_loss = val_error
                if self.multi_gpu:
                    save_model = self.model.module
                else:
                    save_model = self.model
                torch.save(save_model,#.state_dict(),
                           os.path.join(self.model_path, 'best_model.pt'))
                torch.save({
                            'epoch': self.epoch,
                            'model_state_dict': save_model.state_dict(),
                            'optimizer_state_dict': self.optimizer.state_dict(),
                            'loss': loss
                            },
                    os.path.join(self.model_path, 'best_model_state.tar')
                )

                # save irc predictions
                irc_mae_E = irc_mae_F = 0.0
                if irc_generator is not None:
                    outputs = self.validation('irc', irc_generator, irc_steps)
                    irc_mae_E = np.mean(outputs['E_ae'])
                    irc_mae_F = np.mean(outputs['F_ae_masked'])
                    # np.save(os.path.join(self.val_out_path, 'irc_ae_E'), outputs['E_ae'])
                    # np.save(os.path.join(self.val_out_path, 'irc_ae_F'), outputs['F_ae'])
                    # np.save(os.path.join(self.val_out_path, 'irc_pred_E'), outputs['E_pred'])
                    # np.save(os.path.join(self.val_out_path, 'irc_pred_F'), outputs['F_pred'])
                    # np.save(os.path.join(self.val_out_path, 'irc_E'), outputs['E'])
                    # np.save(os.path.join(self.val_out_path, 'irc_F'), outputs['F'])
                    # np.save(os.path.join(self.val_out_path, 'irc_Ei_best'), outputs['Ei'])
                    # np.save(os.path.join(self.val_out_path, 'irc_AM'), outputs['AM'])
                    # np.save(os.path.join(self.val_out_path, 'irc_RM'), outputs['RM'])
                    # np.save(os.path.join(self.val_out_path, 'irc_Ei_epoch%i'%self.epoch), outputs['Ei'])

                # save test predictions
                if test_generator is not None and self.epoch - last_test_epoch >= self.check_test:
                    if self.mode in ["energy/force", "energy"]:
                        outputs = self.validation('test', test_generator, test_steps)
                        test_mae_E = np.mean(outputs['E_ae'])
                        test_mae_F = np.mean(outputs['F_ae_masked'])
                        # np.save(os.path.join(self.val_out_path, 'test_ae_E'), outputs['E_ae'])
                        # np.save(os.path.join(self.val_out_path, 'test_ae_F'), outputs['F_ae'])
                        # np.save(os.path.join(self.val_out_path, 'test_pred_E'), outputs['E_pred'])
                        # np.save(os.path.join(self.val_out_path, 'test_pred_F'), outputs['F_pred'])
                        # np.save(os.path.join(self.val_out_path, 'test_E'), outputs['E'])
                        # np.save(os.path.join(self.val_out_path, 'test_F'), outputs['F'])
                        # np.save(os.path.join(self.val_out_path, 'test_AM'), outputs['AM'])
                        # np.save(os.path.join(self.val_out_path, 'test_RM'), outputs['RM'])
                    elif self.mode == "atomic_properties":
                        outputs = self.validation_atomic_properties('test', "CS", test_generator, test_steps)
                        # torch.save(outputs, os.path.join(self.val_out_path, 'test_results.pkl'))
                        test_error = outputs["RMSE"]
                    last_test_epoch = self.epoch
                    # np.save(os.path.join(self.val_out_path, 'test_Ei_epoch%i'%self.epoch), outputs['Ei'])

            # learning rate decay
            if self.lr_scheduler[0] == 'plateau':
                running_val_loss.append(val_error)
                if len(running_val_loss) > self.lr_scheduler[1]:
                    running_val_loss.pop(0)
                accum_val_loss = np.mean(running_val_loss)
                self.scheduler.step(accum_val_loss)
            elif self.lr_scheduler[0] == 'decay':
                self.scheduler.step()
                accum_val_loss = 0.0

            # checkpoint
            if self.epoch % self.check_log == 0:

                for i, param_group in enumerate(
                        self.scheduler.optimizer.param_groups):
                    old_lr = float(param_group["lr"])

                if self.mode in ["energy/force", "energy"]:
                    self.store_checkpoint({
                        "loss(MSE)": running_loss,
                        'tr_E(MAE)': ae_energy,
                        'tr_F(MAE)': ae_force,
                        'val_E(MAE)': val_mae_E,
                        'val_F(MAE)': val_mae_F,
                        'irc_E(MAE)': irc_mae_E,
                        'irc_F(MAE)': irc_mae_F,
                        'test_E(MAE)': test_mae_E,
                        'test_F(MAE)': test_mae_F,
                        "lr": old_lr,
                        "time": time.time() - t0},
                        steps)
                elif self.mode == "atomic_properties":
                    self.store_checkpoint({
                        "loss(MSE)": running_loss,
                        "tr_err(RMSE)": rmse_ai,
                        "val_err(RMSE)": val_error,
                        "test_err(RMSE)": test_error,
                        "lr": old_lr,
                        "time": time.time() - t0
                    }, steps)

    def log_statistics(self, n_train_data, n_val_data, n_test_data, normalizer, test_energy_hash):
        with open(os.path.join(self.output_path, "stats.txt"), "w") as f:
            f.write("Train data: %d\n" % n_train_data)
            f.write("Val data: %d\n" % n_val_data)
            f.write("Test data: %d\n" % n_test_data)
            f.write("Normalizer: %s\n" % str(normalizer))
            f.write("Test energy hash: %s\n" % test_energy_hash)



import numpy as np
import torch
from torch.utils.data import Dataset

from newtonnet.utils import standardize_batch
from newtonnet.utils import rotate_molecule, euler_rotation_matrix


class BatchDataset(Dataset):
    """
    Parameters
    ----------
    input: dict
        The dictionary of batch data in ndarray format.

    """
    def __init__(self, input, device):

        self.R = torch.tensor(input['R'],
                                  device=device,
                                  # dtype=torch.float64
                              )

        self.Z = torch.tensor(input['Z'],
                                           dtype=torch.long,
                                           device=device)
        self.E = torch.tensor(input['E'],
                                   # dtype=torch.float64,
                                   device=device)
        self.F = torch.tensor(input['F'],
                                   # dtype=torch.float64,
                                   device=device)
        self.AM = torch.tensor(input['AM'],
                                dtype=torch.long,
                                device=device)
        if 'NM' in input and input['NM'] is not None:
            self.N = torch.tensor(input['N'],
                                   dtype=torch.long,
                                   device=device)
            self.NM = torch.tensor(input['NM'],
                                   dtype=torch.long,
                                   device=device)

        # rotation matrix
        self.RM = None
        if 'RM' in input and input['RM'] is not None:
            self.RM = torch.tensor(input['RM'],
                                   # dtype=torch.float64,
                                   device=device)

    def __getitem__(self, index):

        output = dict()
        output['R'] = self.R[index]
        output['Z'] = self.Z[index]
        output['E'] = self.E[index]
        output['F'] = self.F[index]
        output['AM'] = self.AM[index]
        output['N'] = self.N[index]
        output['NM'] = self.NM[index]
        if self.RM is not None:
            output['RM'] = self.RM[index]

        return output

    def __len__(self):
        return self.R.size()[0]


def batch_dataset_converter(input, device):
    result = {}
    result["R"] = torch.tensor(input['R'],
                                  # dtype=torch.float64,
                                   device=device)

    result["Z"] = torch.tensor(input['Z'],
                                           dtype=torch.long,
                                   device=device)
    if "E" in input:
        result["E"] = torch.tensor(input['E'],
                                   dtype=torch.float32,
                                   device=device)
    if "F" in input:
        result["F"] = torch.tensor(input['F'],
                                    dtype=torch.float32,
                                    device=device)
    else:
        result["F"] = None
    if "CS" in input:
        result["CS"] = torch.tensor(input['CS'],
                                    dtype=torch.float32,
                                    device=device)
    if "cs_scaler" in input:
        cs_scaler = input['cs_scaler']
        cs_scaler[cs_scaler == 0] = 1 
        result["cs_scaler"] = torch.tensor(input['cs_scaler'],
                                    dtype=torch.float32,
                                    device=device)

    if "M" in input:
        result["M"] = torch.tensor(input['M'],
                                dtype=torch.long,
                                device=device)
    result["AM"] = torch.tensor(input['AM'],
                                dtype=torch.long,
                                   device=device)
    if 'NM' in input and input['NM'] is not None:
        result["N"] = torch.tensor(input['N'],
                                   dtype=torch.long,
                                   device=device)
        result["NM"] = torch.tensor(input['NM'],
                                   dtype=torch.long,
                                   device=device)
    
    if 'lattice' in input:
        result['lattice'] = torch.tensor(input['lattice'],
                                         device=device)
    if 'D' in input:
        result['D'] = torch.tensor(input['D'],
                                    device=device)

    if 'V' in input:
        result['V'] = torch.tensor(input['V'],
                                    device=device)

    if 'NA' in input:
        result['NA'] = torch.tensor(input['NA'],
                                    dtype=torch.long,
                                    device=device)

    # rotation matrix
    result["RM"] = None
    if 'RM' in input and input['RM'] is not None:
        result["RM"] = torch.tensor(input['RM'],
                                # dtype=torch.float64,
                                   device=device)

    if "labels" in input:
        result["labels"] = input["labels"]
    return result


def extensive_train_loader(data,
                           env_provider=None,
                           batch_size=32,
                           n_rotations=0,
                           freeze_rotations=False,
                           keep_original=True,
                           device=None,
                           shuffle=True,
                           drop_last=False):
    r"""
    The main function to load and iterate data based on the extensive environment provider.

    Parameters
    ----------
    data: dict
        dictionary of arrays with following keys:
            - 'R':positions
            - 'Z':atomic_numbers
            - 'E':energy
            - 'F':forces

    env_provider: ShellProvider
        the instance of combust.data.ExtensiveEnvironment or combust.data.PeriodicEnvironment

    batch_size: int, optional (default: 32)
        The size of output tensors

    n_rotations: int, optional (default: 0)
        Number of times to rotate voxel boxes for data augmentation.
        If zero, the original orientation will be used.

    freeze_rotations: bool, optional (default: False)
        If True rotation angles will be determined and fixed during generation.

    keep_original: bool, optional (default: True)
        If True, the original orientation of data is kept in each epoch

    device: torch.device
        either cpu or gpu (cuda) device.

    shuffle: bool, optional (default: True)
        If ``True``, shuffle the list of file path and batch indices between iterations.

    drop_last: bool, optional (default: False)
        set to ``True`` to drop the last incomplete batch,
        if the dataset size is not divisible by the batch size. If ``False`` and
        the size of dataset is not divisible by the batch size, then the last batch
        will be smaller. (default: ``False``)

    Yields
    -------
    BatchDataset: instance of BatchDataset with the all batch data

    """
    n_data = data['R'].shape[0]  # D
    # n_atoms = data['R'].shape[1]  # A

    # print("number of data w/o augmentation: ", n_data)



    # # shuffle
    # if shuffle:
    #     shuffle_idx = np.arange(n_data)
    #     carts = carts[shuffle_idx]
    #     atomic_numbers = atomic_numbers[shuffle_idx]
    #     energy = energy[shuffle_idx]
    #     forces = forces[shuffle_idx]

    if isinstance(freeze_rotations, list):
        thetas = freeze_rotations
        freeze_rotations = True
    else:
        if freeze_rotations:
            import itertools
            theta_gen = itertools.permutations(np.linspace(-np.pi, np.pi, 6), 3) #todo: linspace size
            thetas = [np.array([0.0, 0.0, 0.0])
                      ]  # index 0 is reserved for the original data (no rotation)
            thetas += [
                np.random.uniform(-np.pi, np.pi, size=3)
                for _ in range(n_rotations)
            ]

    # iterate over data snapshots
    seen_all_data = 0
    while True:

        # iterate over rotations
        for r in range(n_rotations + 1):

            # split by batch size and yield
            data_atom_indices = list(range(n_data))

            if shuffle:
                np.random.shuffle(data_atom_indices)

            split = 0
            while (split + 1) * batch_size <= n_data:
                # Output a batch
                data_batch_idx = data_atom_indices[split *
                                                   batch_size:(split + 1) *
                                                   batch_size]
                data_batch = {k: standardize_batch(data[k][data_batch_idx]) for k in data if k != "labels"}
                if "labels" in data:
                    data_batch["labels"] = [data["labels"][idx] for idx in data_batch_idx]
                # get neighbors
                if env_provider is not None:
                    if 'lattice' in data_batch:
                        neighbors, neighbor_mask, atom_mask, distances, distance_vectors = \
                        env_provider.get_environment(data_batch['R'], data_batch['Z'], data_batch['lattice'])
                    else:
                        neighbors, neighbor_mask, atom_mask, distances, distance_vectors = \
                        env_provider.get_environment(data_batch['R'], data_batch['Z'])

                # rotation
                if r > 0:
                    if freeze_rotations:
                        try:
                            theta = np.array(next(theta_gen))
                        except:
                            theta_gen = itertools.permutations(
                            np.linspace(-np.pi, np.pi, 6), 3)
                            theta = np.array(next(theta_gen))
                    else:
                        theta = np.random.uniform(-np.pi, np.pi, size=3)
                    rot_atoms = rotate_molecule(data_batch['R'], theta=theta)  # B, A, 3
                    if distance_vectors is not None:
                        rot_dist_vec = rotate_molecule(distance_vectors, theta=theta)
                    if 'F' in data:
                        rot_forces = rotate_molecule(data_batch['F'], theta=theta)  # B, A, 3
                else:
                    if not keep_original:
                        theta = np.random.uniform(-np.pi, np.pi, size=3)
                        rot_atoms = rotate_molecule(data_batch['R'], theta=theta)  # B, A, 3
                        if distance_vectors is not None:
                            rot_dist_vec = rotate_molecule(distance_vectors, theta=theta)
                        if 'F' in data:
                            rot_forces = rotate_molecule(data_batch['F'], theta=theta)  # B, A, 3
                    else:
                        theta = np.array([0,0,0])
                        rot_atoms = data_batch['R']
                        if 'F' in data:
                            rot_forces = data_batch['F']
                        if distance_vectors is not None:
                            rot_dist_vec = distance_vectors

                RM = np.tile(theta, (rot_atoms.shape[0],1))

                if env_provider is None:
                    N = None
                    NM = None
                    Z = data_batch['Z']
                    AM = np.zeros_like(Z)
                    AM[Z != 0] = 1
                else:
                    N = neighbors
                    NM = neighbor_mask
                    AM = atom_mask

                batch_dataset = {k:v for k,v in data_batch.items()}
                batch_dataset.update({'R': rot_atoms,  # B,A,3
                                      'N': N,     # B,A,A-1
                                      'NM': NM,   # B,A,A-1
                                      'AM': AM,   # B,A
                                      'RM': RM    # B,3  rotation angles only
                                      })
                if 'F' in data:
                    batch_dataset.update({'F': rot_forces})

                if distances is not None and distance_vectors is not None:
                    batch_dataset.update({'D': distances,
                                          'V': rot_dist_vec})

                # batch_dataset = {
                #     'R': rot_atoms,   # B,A,3
                #     'Z': data_batch['Z'], # B,A
                #     'E': data_batch['E'], # B,1
                #     'F': rot_forces if 'F' in data else None,    # B,A,3
                #     'N': N,     # B,A,A-1
                #     'NM': NM,   # B,A,A-1
                #     'AM': AM,   # B,A
                #     'RM': RM    # B,3  rotation angles only
                # }
                # batch_dataset = BatchDataset(batch_dataset, device=device)
                batch_dataset = batch_dataset_converter(batch_dataset, device)
                yield batch_dataset
                split += 1

            # Deal with the part smaller than a batch_size
            left_len = n_data % batch_size
            if left_len != 0 and drop_last:
                continue

            elif left_len != 0 and not drop_last:
                left_idx = data_atom_indices[split * batch_size:]
                data_batch = {k: standardize_batch(data[k][left_idx]) for k in data for k in data if k != "labels"}
                if "labels" in data:
                    data_batch["labels"] = [data["labels"][idx] for idx in left_idx]
                # get neighbors
                if env_provider is not None:
                    if 'lattice' in data:
                        neighbors, neighbor_mask, atom_mask, distances, distance_vectors = \
                        env_provider.get_environment(data_batch['R'], data_batch['Z'], data_batch['lattice'])
                    else:
                        neighbors, neighbor_mask, atom_mask, distances, distance_vectors = \
                        env_provider.get_environment(data_batch['R'], data_batch['Z'])

                # rotation
                if r > 0:
                    if freeze_rotations:
                        try:
                            theta = np.array(next(theta_gen))
                        except:
                            theta_gen = itertools.permutations(
                            np.linspace(-np.pi, np.pi, 6), 3)
                            theta = np.array(next(theta_gen))
                    else:
                        theta = np.random.uniform(-np.pi, np.pi, size=3)
                    rot_atoms = rotate_molecule(data_batch['R'], theta=theta)  # B, A, 3
                    if distance_vectors is not None:
                        rot_dist_vec = rotate_molecule(distance_vectors, theta=theta)
                    if 'F' in data:
                        rot_forces = rotate_molecule(data_batch['F'], theta=theta)  # B, A, 3
                else:
                    if not keep_original:
                        theta = np.random.uniform(-np.pi, np.pi, size=3)
                        rot_atoms = rotate_molecule(data_batch['R'], theta=theta)  # B, A, 3
                        if distance_vectors is not None:
                            rot_dist_vec = rotate_molecule(distance_vectors, theta=theta)
                        if 'F' in data:
                            rot_forces = rotate_molecule(data_batch['F'], theta=theta)  # B, A, 3
                    else:
                        theta = np.array([0,0,0])
                        rot_atoms = data_batch['R']
                        if 'F' in data:
                            rot_forces = data_batch['F']
                        if distance_vectors is not None:
                            rot_dist_vec = distance_vectors

                RM = np.tile(theta, (rot_atoms.shape[0],1))

                if env_provider is None:
                    N = None
                    NM = None
                    Z = data_batch['Z']
                    AM = np.zeros_like(Z)
                    AM[Z != 0] = 1
                else:
                    N = neighbors
                    NM = neighbor_mask
                    AM = atom_mask

                batch_dataset = {k:v for k,v in data_batch.items()}
                batch_dataset.update({'R': rot_atoms,  # B,A,3
                                      'N': N,     # B,A,A-1
                                      'NM': NM,   # B,A,A-1
                                      'AM': AM,   # B,A
                                      'RM': RM    # B,3  rotation angles only
                                      })

                if 'F' in data:
                    batch_dataset.update({'F': rot_forces})

                if distances is not None and distance_vectors is not None:
                    batch_dataset.update({'D': distances,
                                          'V': rot_dist_vec})

                # batch_dataset = {
                #     'R': rot_atoms,
                #     'Z': data_batch['Z'],
                #     'E': data_batch['E'],
                #     'F': rot_forces if 'F' in data else None,
                #     'N': N,
                #     'NM': NM,
                #     'AM': AM,
                #     'RM': RM
                # }

                # batch_dataset = BatchDataset(batch_dataset, device)
                batch_dataset = batch_dataset_converter(batch_dataset, device)
                yield batch_dataset

            seen_all_data += 1
            # print('\n# trained on entire data: %i (# rotation: %i)\n'%(seen_all_data, (n_rotations+1)))


def extensive_loader_rotwise(
                           data,
                           env_provider=None,
                           batch_size=32,
                           n_rotations=0,
                           freeze_rotations=False,
                           keep_original=True,
                           device=None,
                           shuffle=True,
                           drop_last=False):
    r"""
    The main function to load and iterate data based on the extensive environment provider.
    In this implementation we keep track of all rotations of a data point (compared to `extensive_train_loader`)

    Parameters
    ----------
    data: dict
        dictionary of arrays with following keys:
            - 'R':positions
            - 'Z':atomic_numbers
            - 'E':energy
            - 'F':forces

    env_provider: ShellProvider
        the instance of combust.data.ExtensiveEnvironment

    batch_size: int, optional (default: 32)
        The size of output tensors

    n_rotations: int, optional (default: 0)
        Number of times to rotate voxel boxes for data augmentation.
        If zero, the original orientation will be used.

    freeze_rotations: bool, optional (default: False)
        If True rotation angles will be determined and fixed during generation.

    keep_original: bool, optional (default: True)
        If True, the original orientation of data is kept in each epoch

    device: torch.device
        either cpu or gpu (cuda) device.

    shuffle: bool, optional (default: True)
        If ``True``, shuffle the list of file path and batch indices between iterations.

    drop_last: bool, optional (default: False)
        set to ``True`` to drop the last incomplete batch,
        if the dataset size is not divisible by the batch size. If ``False`` and
        the size of dataset is not divisible by the batch size, then the last batch
        will be smaller. (default: ``False``)

    Yields
    -------
    BatchDataset: instance of BatchDataset with the all batch data

    """
    n_data = data['R'].shape[0]  # D
    n_atoms = data['R'].shape[1]  # A

    # print("number of data w/o augmentation: ", n_data)

    # get neighbors
    if env_provider is not None:
        neighbors, neighbor_mask, atom_mask = env_provider.get_environment(data['R'], data['Z'])

    # # shuffle
    # if shuffle:
    #     shuffle_idx = np.arange(n_data)
    #     carts = carts[shuffle_idx]
    #     atomic_numbers = atomic_numbers[shuffle_idx]
    #     energy = energy[shuffle_idx]
    #     forces = forces[shuffle_idx]

    if isinstance(freeze_rotations, list):
        thetas = freeze_rotations
        freeze_rotations = True
    else:
        if freeze_rotations:
            thetas = [np.array([0.0, 0.0, 0.0])
                      ]  # index 0 is reserved for the original data (no rotation)
            thetas += [
                np.random.uniform(-np.pi, np.pi, size=3)
                for _ in range(n_rotations)
            ]

    # iterate over data snapshots
    seen_all_data = 0
    while True:

        # split by batch size and yield
        data_atom_indices = list(range(n_data))

        if shuffle:
            np.random.shuffle(data_atom_indices)

        split = 0
        while (split + 1) * batch_size <= n_data:
            # Output a batch
            data_batch_idx = data_atom_indices[split *
                                               batch_size:(split + 1) *
                                               batch_size]

            # rotation
            if not freeze_rotations:
                if keep_original:
                    thetas = np.random.uniform(-np.pi, np.pi, size=(n_rotations, 3))
                    thetas = [np.array([0,0,0])] + list(thetas)
                else:
                    thetas = np.random.uniform(-np.pi, np.pi, size=(n_rotations+1, 3))

            # stack all rotations
            Rs=[]; Fs=[]; RMs=[]
            for theta in thetas:
                rot_atoms = rotate_molecule(data['R'][data_batch_idx], theta=theta)  # B, A, 3
                # rot_forces = rotate_molecule(data['F'][data_batch_idx], theta=theta)  # B, A, 3
                rot_matrix = euler_rotation_matrix(theta)   # 3,3

                Rs.append(rot_atoms)
                # Fs.append(rot_forces)
                RMs.append(rot_matrix)

            rot_atoms = np.stack(Rs, axis=1)  # B,n_rot,A,3
            rsize = rot_atoms.shape
            rot_atoms = rot_atoms.reshape(rsize[0]*rsize[1], rsize[2], rsize[3]) # B*n_rot, A, 3
            # rot_forces= np.stack(Fs, axis=1)  # B,n_rot,A,3
            # rot_forces= rot_forces.reshape(rsize[0]*rsize[1], rsize[2], rsize[3]) # B*n_rot, A, 3

            RM = np.stack(RMs, axis=0)  # n_rot,3,3
            RM = np.tile(RM, (rsize[0], 1, 1, 1))    # B, n_rot, 3, 3

            if env_provider is None:
                N = None
                NM = None
                E = data['E'][data_batch_idx]  # B,1
                # E = np.tile(E, (1, n_rotations + 1)).reshape(E.shape[0] * (n_rotations + 1), E.shape[1])  # B*n_rot,1

                Z = data['Z'][data_batch_idx]  # B,A
                Z = np.tile(Z, (1,n_rotations+1)).reshape(Z.shape[0],n_rotations+1,Z.shape[1]) # B,n_rot,A

                AM = np.zeros_like(Z)   # B,n_rot,A
                AM[Z != 0] = 1
                RM = RM

            else:
                E = data['E'][data_batch_idx]  # B,1
                # E = np.tile(E, (1, n_rotations + 1)).reshape(E.shape[0] * (n_rotations + 1), E.shape[1])  # B*n_rot,1

                Z = data['Z'][data_batch_idx]  # B,A
                Z = np.tile(Z, (1,n_rotations+1)).reshape(Z.shape[0]*(n_rotations+1),Z.shape[1]) # B*n_rot,A

                N = neighbors[data_batch_idx]  # B,A,A-1
                N = np.tile(N, (1,n_rotations+1,1)).reshape(N.shape[0]*(n_rotations+1),N.shape[1], N.shape[2]) # B*n_rot,A,A-1

                NM = neighbor_mask[data_batch_idx]  # B,A,A-1
                NM = np.tile(NM, (1,n_rotations+1,1)).reshape(NM.shape[0]*(n_rotations+1),NM.shape[1], NM.shape[2]) # B*n_rot,A,A-1

                AM = atom_mask[data_batch_idx]   # B,A
                AM = np.tile(AM, (1,n_rotations+1)).reshape(AM.shape[0]*(n_rotations+1),AM.shape[1]) # B*n_rot,A

                # rotation mask
                # B = rsize[0]; n_rot=rsize[1]
                # RotM = np.eye(B)
                # RotM = np.tile(RotM, (1, n_rot)).reshape(B*n_rot, B)

                RM = RM

            batch_dataset = {
                'R': rot_atoms,   # B,A,3
                'Z': Z, # B,A
                'E': E, # B,1
                'F': data['F'][data_batch_idx],    # B,A,3
                'N': N,     # B,A,A-1
                'NM': NM,   # B,A,A-1
                'AM': AM,   # B,A
                'RM': RM
            }

            # batch_dataset = BatchDataset(batch_dataset, device=device)
            batch_dataset = batch_dataset_converter(batch_dataset, device)
            yield batch_dataset
            split += 1

        # Deal with the part smaller than a batch_size
        left_len = n_data % batch_size
        if left_len != 0 and drop_last:
            continue

        elif left_len != 0 and not drop_last:
            left_idx = data_atom_indices[split * batch_size:]

            # rotation
            if not freeze_rotations:
                if keep_original:
                    thetas = np.random.uniform(-np.pi, np.pi, size=(n_rotations, 3))
                    thetas = [np.array([0,0,0])] + list(thetas)
                else:
                    thetas = np.random.uniform(-np.pi, np.pi, size=(n_rotations+1, 3))

            # stack all rotations
            Rs=[]; Fs=[]; RMs=[]
            for theta in thetas:
                rot_atoms = rotate_molecule(data['R'][left_idx], theta=theta)  # B, A, 3
                # rot_forces = rotate_molecule(data['F'][left_idx], theta=theta)  # B, A, 3
                rot_matrix = euler_rotation_matrix(theta)   # 3,3

                Rs.append(rot_atoms)
                # Fs.append(rot_forces)
                RMs.append(rot_matrix)

            rot_atoms = np.stack(Rs, axis=1)  # B,n_rot,A,3
            rsize = rot_atoms.shape
            rot_atoms = rot_atoms.reshape(rsize[0]*rsize[1], rsize[2], rsize[3]) # B*n_rot, A, 3
            # rot_forces= np.stack(Fs, axis=1)  # B,n_rot,A,3
            # rot_forces= rot_forces.reshape(rsize[0]*rsize[1], rsize[2], rsize[3]) # B*n_rot, A, 3

            RM = np.stack(RMs, axis=0)  # n_rot,3,3
            RM = np.tile(RM, (rsize[0], 1, 1, 1))    # B, n_rot, 3, 3

            if env_provider is None:
                N = None
                NM = None

                E = data['E'][left_idx]  # B,1
                # E = np.tile(E, (1,n_rotations+1)).reshape(E.shape[0]*(n_rotations+1),E.shape[1]) # B*n_rot,1

                Z = data['Z'][left_idx]
                Z = np.tile(Z, (1,n_rotations+1)).reshape(Z.shape[0]*(n_rotations+1),Z.shape[1]) # B*n_rot,A

                AM = np.zeros_like(Z)
                AM[Z != 0] = 1
                RM = RM
            else:
                E = data['E'][left_idx]  # B,1
                # E = np.tile(E, (1,n_rotations+1)).reshape(E.shape[0]*(n_rotations+1),E.shape[1]) # B*n_rot,1

                Z = data['Z'][left_idx]  # B,A
                Z = np.tile(Z, (1,n_rotations+1)).reshape(Z.shape[0]*(n_rotations+1),Z.shape[1]) # B*n_rot,A

                N = neighbors[left_idx]  # B,A,A-1
                N = np.tile(N, (1,n_rotations+1,1)).reshape(N.shape[0]*(n_rotations+1),N.shape[1], N.shape[2]) # B*n_rot,A,A-1

                NM = neighbor_mask[left_idx]  # B,A,A-1
                NM = np.tile(NM, (1,n_rotations+1,1)).reshape(NM.shape[0]*(n_rotations+1),NM.shape[1], NM.shape[2]) # B*n_rot,A,A-1

                AM = atom_mask[left_idx]   # B,A
                AM = np.tile(AM, (1,n_rotations+1)).reshape(AM.shape[0]*(n_rotations+1),AM.shape[1]) # B*n_rot,A

                RM = RM

            batch_dataset = {
                'R': rot_atoms,
                'Z': Z,
                'E': E,
                'F': data['F'][left_idx],
                'N': N,
                'NM': NM,
                'AM': AM,
                'RM': RM
            }

            # batch_dataset = BatchDataset(batch_dataset, device)
            batch_dataset = batch_dataset_converter(batch_dataset, device)
            yield batch_dataset

        seen_all_data += 1
        # print('\n# trained on entire data: %i (# rotation: %i)\n'%(seen_all_data, (n_rotations+1)))

def extensive_voxel_loader(data,
                           env_provider=None,
                           batch_size=32,
                           device=None,
                           shuffle=True,
                           drop_last=False):
    r"""
    The main function to load and iterate data based on the extensive environment provider.

    Parameters
    ----------
    data: dict
        dictionary of arrays with following keys:
            - 'R':positions
            - 'Z':atomic_numbers
            - 'E':energy
            - 'F':forces

    env_provider: ShellProvider
        the instance of combust.data.ExtensiveEnvironment

    batch_size: int, optional (default: 32)
        The size of output tensors

    device: torch.device
        either cpu or gpu (cuda) device.

    shuffle: bool, optional (default: True)
        If ``True``, shuffle the list of file path and batch indices between iterations.

    drop_last: bool, optional (default: False)
        set to ``True`` to drop the last incomplete batch,
        if the dataset size is not divisible by the batch size. If ``False`` and
        the size of dataset is not divisible by the batch size, then the last batch
        will be smaller. (default: ``False``)

    Yields
    -------
    BatchDataset: instance of BatchDataset with the all batch data

    """
    n_data = data['R'].shape[0]  # D
    n_atoms = data['R'].shape[1]  # A

    # print("number of data w/o augmentation: ", n_data)

    # get neighbors
    if env_provider is not None:
        neighbors, neighbor_mask, atom_mask = env_provider.get_environment(data['R'], data['Z'])

    # iterate over data snapshots
    seen_all_data = 0
    while True:
        # split by batch size and yield
        data_atom_indices = list(range(n_data))

        if shuffle:
            np.random.shuffle(data_atom_indices)

        split = 0
        while (split + 1) * batch_size <= n_data:
            # Output a batch
            data_batch_idx = data_atom_indices[split *
                                               batch_size:(split + 1) *
                                               batch_size]

            if env_provider is None:
                N = None
                NM = None
                Z = data['Z'][data_batch_idx]
                AM = np.zeros_like(Z)
                AM[Z != 0] = 1
                thetas = np.random.uniform(-np.pi, np.pi, size=(Z.shape[0],3))
                RM = np.array([euler_rotation_matrix(t) for t in thetas])
            else:
                N = neighbors[data_batch_idx]
                NM = neighbor_mask[data_batch_idx]
                AM = atom_mask[data_batch_idx]
                RM = None


            batch_dataset = {
                'R': data['R'][data_batch_idx],
                'Z': data['Z'][data_batch_idx],
                'E': data['E'][data_batch_idx],
                'F': data['F'][data_batch_idx],
                'N': N,
                'NM': NM,
                'AM': AM,
                'RM': RM
            }
            # batch_dataset = BatchDataset(batch_dataset, device=device)
            batch_dataset = batch_dataset_converter(batch_dataset, device)
            yield batch_dataset
            split += 1

        # Deal with the part smaller than a batch_size
        left_len = n_data % batch_size
        if left_len != 0 and drop_last:
            continue

        elif left_len != 0 and not drop_last:

            left_idx = data_atom_indices[split * batch_size:]

            if env_provider is None:
                N = None
                NM = None
                Z = data['Z'][left_idx]
                AM = np.zeros_like(Z)
                AM[Z != 0] = 1
                thetas = np.random.uniform(-np.pi, np.pi, size=(Z.shape[0],3))
                RM = np.array([euler_rotation_matrix(t) for t in thetas])
            else:
                N = neighbors[left_idx]
                NM = neighbor_mask[left_idx]
                AM = atom_mask[left_idx]
                RM = None

            batch_dataset = {
                'R': data['R'][left_idx],
                'Z': data['Z'][left_idx],
                'E': data['E'][left_idx],
                'F': data['F'][left_idx],
                'N': N,
                'NM': NM,
                'AM': AM,
                'RM': RM
            }

            # batch_dataset = BatchDataset(batch_dataset, device)
            batch_dataset = batch_dataset_converter(batch_dataset, device)
            yield batch_dataset

            seen_all_data += 1
            # print('\n# trained on entire data: %i (# rotation: %i)\n'%(seen_all_data, (n_rotations+1)))





import numpy as np

import torch
from torch import nn


class RadialBesselLayer(nn.Module):
    r"""Radial Bessel functions based on the work by DimeNet: https://github.com/klicperajo/dimenet

    Args:
        n_radials (int, optional): total number of radial functions, :math:`N_g`.
        cutoff (float, optional): cutoff, :math:`\mu_{r_c}`

    """

    def __init__(
        self, n_radial=16, cutoff=5.0, device=None
    ):
        super(RadialBesselLayer, self).__init__()
        self.inv_cutoff = 1/cutoff
        self.frequencies = nn.Parameter(torch.tensor(np.arange(1,n_radial+1) * np.pi, device=device), requires_grad=False)

    def forward(self, distances):
        """Compute smeared-gaussian distance values.

        Args:
            distances (torch.Tensor): interatomic distance values of
                (N_b x N_at x N_nbh) shape.

        Returns:
            torch.Tensor: layer output of (N_b x N_at x N_nbh x N_g) shape.

        """
        d_scaled = distances * self.inv_cutoff
        d_scaled = d_scaled.unsqueeze(-1)
        out = torch.sin(self.frequencies * d_scaled)

        return out



import os
import numpy as np
import warnings
from collections import defaultdict
from numpy.lib.function_base import append
from sklearn.utils import random
from sklearn.utils.random import sample_without_replacement
from sklearn.model_selection import train_test_split

from newtonnet.data import ExtensiveEnvironment, PeriodicEnvironment
from newtonnet.data import extensive_train_loader, extensive_loader_rotwise

from ase.io import iread
import math
import pickle

def concat_listofdicts(listofdicts, axis=0):
    """

    Parameters
    ----------
    listofdicts: list
        values must be 2d arrays
    axis: int

    Returns
    -------
    dict

    """
    data = dict()
    for k in listofdicts[0].keys():
        data[k] = np.concatenate([d[k] for d in listofdicts], axis=axis)

    return data


def split(data, train_size, test_size, val_size, random_states=90, stratify=None):
    """

    Parameters
    ----------
    data: dict
    train_size: int
    test_size
    val_size
    random_states
    stratify: None or labels

    Returns
    -------
    dict: train data
    dict: val data
    dict: test data

    """

    tr_ind, val_ind = train_test_split(list(range(data['R'].shape[0])),
                                      test_size=val_size,
                                      random_state=random_states,
                                      stratify=stratify)

    if stratify is not None:
        stratify_new = stratify[tr_ind]
    else:
        stratify_new = None

    tr_ind, te_ind = train_test_split(tr_ind,
                                       test_size=test_size,
                                       train_size=train_size,
                                       random_state=random_states,
                                       stratify=stratify_new)

    train = dict()
    val = dict()
    test = dict()
    for key in data:
        train[key] = data[key][tr_ind]
        val[key] = data[key][val_ind]
        test[key] = data[key][te_ind]

    if stratify is not None:
        train['L'] = stratify[tr_ind]
        val['L'] = stratify[val_ind]
        test['L'] = stratify[te_ind]

    return train, val, test


def h2_reaction(reaction_number, settings, all_sets):
    """

    Parameters
    ----------
    reaction_number: int

    settings: dict
        dict of yaml file

    all_sets: dict

    Returns
    -------

    """

    dir_path = settings['data']['root']

    # file name prefix
    if reaction_number < 10:
        pre = '0%i'%reaction_number
    elif reaction_number >= 10:
        pre = '%i'%reaction_number
    # elif reaction_number == 6:
    #     pre = ['0%ia_irc.npz' % reaction_number, '0%ib_irc.npz' % reaction_number]
    # elif reaction_number == 12:
    #     pre = ['%ia_irc.npz' % reaction_number, '%ib_irc.npz' % reaction_number]

    # read npz files
    aimd = nm = irc = None
    aimd_path = os.path.join(dir_path, '%s_aimd.npz'%pre)
    if os.path.exists(aimd_path):
        aimd = dict(np.load(aimd_path))
    nm_path = os.path.join(dir_path, '%s_nm.npz'%pre)
    if os.path.exists(nm_path):
        nm = dict(np.load(nm_path))
    irc_path = os.path.join(dir_path, '%s_irc.npz'%pre)
    if os.path.exists(irc_path):
        irc = dict(np.load(irc_path))

    # merge aimd and normal mode data
    if settings['data']['normal_mode'] and nm is not None:
        data = dict()
        n_nm = min(settings['data']['size_nmode_max'], nm['R'].shape[0])
        nm_select = sample_without_replacement(nm['R'].shape[0],
                                               n_nm,
                                               random_state=settings['data']['random_states'])
        if aimd is not None:
            for k in aimd.keys():
                data[k] = np.concatenate([aimd[k], nm[k][nm_select]], axis=0)

            assert data['R'].shape[0] == (aimd['R'].shape[0]+n_nm)
        else:
            data = None
            warnings.warn('both AIMD and normal mode data for reaction# %i are missing.'%reaction_number)

    elif aimd is not None:
        data = aimd

    else:
        data = None
        warnings.warn('both AIMD and normal mode data for reaction# %i are missing.'%reaction_number)

    if settings['data']['cgem']:
        assert data['E'].shape == data['CE'].shape
        assert data['F'].shape == data['CF'].shape
        data['E'] = data['E'] - data['CE']
        data['F'] = data['F'] - data['CF']
        irc['E'] = irc['E'] - irc['CE']
        irc['F'] = irc['F'] - irc['CF']

    train_size = settings['data']['trsize_perrxn_max']
    if train_size == -1:
        train_size = None  # to select all remaining data in each reaction

    if data is not None:
        dtrain, dval, dtest = split(data,
                                    train_size=train_size,
                                    test_size=settings['data']['test_size'],
                                    val_size=settings['data']['val_size'],
                                    random_states=settings['data']['random_states'],
                                    stratify=None)
    else:
        dtrain, dval, dtest = None, None, None

    # compile data sets
    all_sets['train'].append(dtrain)
    all_sets['val'].append(dval)
    all_sets['test'].append(dtest)
    all_sets['irc'].append(irc)

    return all_sets


def parse_h2_reaction(settings, device):
    """

    Parameters
    ----------
    settings: instance of yaml file
    device: torch devices

    Returns
    -------
    generator: train, val, irc, test generators, respectively
    int: n_steps for train, val, irc, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data

    """

    # list of reaction number(s)
    reaction_number = settings['data']['reaction']

    if isinstance(reaction_number, int):
        reaction_number = [reaction_number]

    # compile dictionary of train, test, val, and irc data
    all_sets = defaultdict(list)
    for rxn_n in reaction_number:
        all_sets = h2_reaction(rxn_n, settings, all_sets)

    dtrain = concat_listofdicts(all_sets['train'], axis=0)
    dval = concat_listofdicts(all_sets['val'], axis=0)
    dtest = concat_listofdicts(all_sets['test'], axis=0)
    irc = concat_listofdicts(all_sets['irc'], axis=0)

    # final down-sampling of training data
    n_train = settings['data']['train_size']
    if n_train == -1:
        n_train = dtrain['R'].shape[0]

    n_train = min(n_train, dtrain['R'].shape[0])
    n_select = sample_without_replacement(dtrain['R'].shape[0],
                                           n_train,
                                           random_state=settings['data']['random_states'])
    for k in dtrain.keys():
        dtrain[k] = dtrain[k][n_select]

    normalizer = (dtrain['E'].mean(), dtrain['E'].std())

    n_tr_data = dtrain['R'].shape[0]
    n_val_data = dval['R'].shape[0]
    n_irc_data = irc['R'].shape[0]
    n_test_data = dtest['R'].shape[0]
    print("# data (train,val,test,irc): %i, %i, %i, %i"%(n_tr_data,n_val_data,n_test_data,n_irc_data))

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # freeze rotatios
    # Todo: it seems that we don't need separated tr and val anymore
    # Todo: consider keep_original scenario in the code
    # if settings['training']['tr_frz_rot']:
    #     if settings['training']['saved_angle_path']:
    #         tr_fra_rot = list(np.load(settings['training']['saved_angle_path']))[:tr_rotations+1]
    #     tr_frz_rot = (np.random.uniform(-np.pi, np.pi, size=3)
    #                   for _ in range(tr_rotations+1))
    #     val_frz_rot = tr_frz_rot
    # else:
    #     tr_frz_rot = settings['training']['tr_frz_rot']
    #     val_frz_rot = settings['training']['val_frz_rot']

    # generators
    project = settings['general']['driver']
    if project not in ['voxel_cart_rotwise.py']:
        # steps
        tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
        val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
        irc_steps = int(np.ceil(n_irc_data / val_batch_size)) * (val_rotations + 1)
        test_steps= int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

        env = ExtensiveEnvironment()

        train_gen = extensive_train_loader(data=dtrain,
                                           env_provider=env,
                                           batch_size=tr_batch_size,
                                           n_rotations=tr_rotations,
                                           freeze_rotations=settings['training']['tr_frz_rot'],
                                           keep_original=settings['training']['tr_keep_original'],
                                           device=device,
                                           shuffle=settings['training']['shuffle'],
                                           drop_last=settings['training']['drop_last'])

        val_gen = extensive_train_loader(data=dval,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=settings['training']['shuffle'],
                                         drop_last=settings['training']['drop_last'])

        irc_gen = extensive_train_loader(data=irc,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=False,
                                         drop_last=False)

        test_gen = extensive_train_loader(data=dtest,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=False,
                                         drop_last=False)

        return train_gen, val_gen, irc_gen, test_gen, tr_steps, val_steps, irc_steps, test_steps, normalizer

    else:

        tr_steps = int(np.ceil(n_tr_data / tr_batch_size))
        val_steps = int(np.ceil(n_val_data / val_batch_size))
        irc_steps = int(np.ceil(n_irc_data / val_batch_size))
        test_steps = int(np.ceil(n_test_data / val_batch_size))

        env = ExtensiveEnvironment()

        train_gen = extensive_loader_rotwise(data=dtrain,
                                           env_provider=env,
                                           batch_size=tr_batch_size,
                                           n_rotations=tr_rotations,
                                           freeze_rotations=settings['training']['tr_frz_rot'],
                                           keep_original=settings['training']['tr_keep_original'],
                                           device=device,
                                           shuffle=settings['training']['shuffle'],
                                           drop_last=settings['training']['drop_last'])

        val_gen = extensive_loader_rotwise(data=dval,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=settings['training']['shuffle'],
                                         drop_last=settings['training']['drop_last'])

        irc_gen = extensive_loader_rotwise(data=irc,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=False,
                                         drop_last=False)

        test_gen = extensive_loader_rotwise(data=dtest,
                                         env_provider=env,
                                         batch_size=val_batch_size,
                                         n_rotations=val_rotations,
                                         freeze_rotations=settings['training']['val_frz_rot'],
                                         keep_original=settings['training']['val_keep_original'],
                                         device=device,
                                         shuffle=False,
                                         drop_last=False)

        return train_gen, val_gen, irc_gen, test_gen, tr_steps, val_steps, irc_steps, test_steps, normalizer


def parse_nmr_data(settings, device, test_only=False):
    '''
    parse and load NMR data from tripeptides/3d fragmentations of SPARTA+/SHIFTX2 dataset with NMR chemical shifts calculated with HF/6-31g* or experimental chemical shifts

    Parameters
    ----------
    settings: instance of yaml file
    device: list
        list of torch devices
    test_only: boolean indicator for whether only read test data

    generator: train, val, test generators, respectively
    int: n_steps for train, val, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data
    '''
    # meta data
    root_folder = settings['data']['root']

    test_proportion = settings['data']['test_proportion']
    val_proportion = settings['data']['val_proportion']
    shift_types = settings['data']['shift_types']

    dtrain = {'R':[], 'Z':[], 'N':[], 'CS':[], "M": []}
    dtest = {'R':[], 'Z':[], 'N':[], 'CS':[], "M": [], "labels": []}

    pbc = settings["data"].get("pbc", False)
    if pbc:
        dtrain["lattice"] = []
        dtest["lattice"] = []

    if test_only:
        test_path = settings['data']['test']
        test_proteins = os.listdir(test_path)
    else:
        if settings['data']['test']:
            test_path = settings['data']['test']
            train_proteins = os.listdir(root_folder)
            test_proteins = os.listdir(test_path)
        else:
            test_path = root_folder
            train_proteins, test_proteins = train_test_split(os.listdir(root_folder),
                        test_size=test_proportion,
                        random_state=settings['data']['random_states'])
        # Load train data
        for protein in train_proteins:
            with open(os.path.join(root_folder, protein), "rb") as f:
                data = pickle.load(f)
            for residue in data:
                if residue['R'].shape[1] > settings['data']['max_natom_cutoff']:
                    continue
                dtrain['R'].append(residue['R'])
                dtrain['Z'].append(residue['Z'])
                dtrain['N'].append([len(residue['R'])])
                shift_type_filter = np.array([item in shift_types for item in residue['Z']])
                dtrain['CS'].append(np.nan_to_num(residue['CS']))
                dtrain['M'].append(residue['M'] & shift_type_filter)
                if pbc:
                    dtrain['lattice'].append(residue['lattice'])


    # Load test data
    for protein in test_proteins:
        with open(os.path.join(test_path, protein), "rb") as f:
            data = pickle.load(f)
        for residue in data:
            if residue['R'].shape[0] > settings['data']['max_natom_cutoff']:
                continue
            dtest['R'].append(residue['R'])
            dtest['Z'].append(residue['Z'])
            dtest['N'].append([len(residue['R'])])
            shift_type_filter = np.array([item in shift_types for item in residue['Z']])
            dtest['CS'].append(np.nan_to_num(residue['CS']))
            dtest['M'].append(residue['M'] & shift_type_filter)
            dtest["labels"].append(residue["meta"])
            if pbc:
                dtest['lattice'].append(residue['lattice'])

    if test_only:
        n_tr_data = n_val_data = 0
        n_test_data = len(dtest['R'])
        for k in dtest:
            dtest[k] = np.array(dtest[k])
        print("test data size: %d"%n_test_data)
    else:
        # Further split train into train/validation
        train_val_indices = list(range(len(dtrain['R'])))
        train_idx, val_idx = train_test_split(train_val_indices, test_size=val_proportion, random_state=settings['data']['random_states'])
        data = dtrain
        dtrain = {}
        dval = {}
        for k in data:
            dtrain[k] = np.array(data[k])[train_idx]
            dval[k] = np.array(data[k])[val_idx]
            dtest[k] = np.array(dtest[k])

        n_tr_data = len(dtrain['R'])
        n_val_data = len(dval['R'])
        n_test_data = len(dtest['R'])
        print("data size: (train,val,test): %i, %i, %i"%(n_tr_data,n_val_data,n_test_data))

        # extract mean and standard deviation of chemical shifts for different atom types
        normalizers = {}
        # normalized_CS = dtrain['CS'].copy()
        for z in shift_types:
            atom_cs = []
            for i in range(n_tr_data):
                atom_cs.extend(dtrain['CS'][i][dtrain['Z'][i] == z])
            normalizers[z] = (np.nanmean(atom_cs), np.nanstd(atom_cs))
        #     for i in range(n_tr_data):
        #         # normalize data for a balanced multi-target training
        #         normalized_CS[i][dtrain['Z'][i] == z] -= normalizers[z][0]
        #         normalized_CS[i][dtrain['Z'][i] == z] /= normalizers[z][1]
        # dtrain['CS_normalized'] = normalized_CS

        atomic_cs_scalers = []
        for i in range(n_tr_data):
            scaler = np.ones_like(dtrain['Z'][i], dtype=float)
            for z in shift_types:
                scaler[dtrain['Z'][i] == z] = normalizers[z][1]
            atomic_cs_scalers.append(scaler)
        dtrain['cs_scaler'] = np.array(atomic_cs_scalers)

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # generators
    me = settings['general']['driver']

    # steps
    tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
    val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
    test_steps= int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

    if settings['data']['pbc']:
        env = PeriodicEnvironment(cutoff=settings['data']['cutoff'] * 1.1)
    else:
        env = ExtensiveEnvironment()

    test_gen = extensive_train_loader(data=dtest,
                                        env_provider=env,
                                        batch_size=val_batch_size,
                                        n_rotations=val_rotations,
                                        freeze_rotations=settings['training']['val_frz_rot'],
                                        keep_original=settings['training']['val_keep_original'],
                                        device=device,
                                        shuffle=False,
                                        drop_last=False)

    if test_only:
        return test_gen, test_steps
    else:
        train_gen = extensive_train_loader(data=dtrain,
                                            env_provider=env,
                                            batch_size=tr_batch_size,
                                            n_rotations=tr_rotations,
                                            freeze_rotations=settings['training']['tr_frz_rot'],
                                            keep_original=settings['training']['tr_keep_original'],
                                            device=device,
                                            shuffle=settings['training']['shuffle'],
                                            drop_last=settings['training']['drop_last'])

        val_gen = extensive_train_loader(data=dval,
                                            env_provider=env,
                                            batch_size=val_batch_size,
                                            n_rotations=val_rotations,
                                            freeze_rotations=settings['training']['val_frz_rot'],
                                            keep_original=settings['training']['val_keep_original'],
                                            device=device,
                                            shuffle=settings['training']['shuffle'],
                                            drop_last=settings['training']['drop_last'])




        return train_gen, val_gen, test_gen, tr_steps, val_steps, test_steps, normalizers

def parse_ani_data(settings, device):
    """
    parse and load the ANI-1 dataset with splitting of train, validation and test
    dataset is expected to be in original .h5 format
    (https://figshare.com/collections/_/3846712)
    energy units: Hartree (convert_unit=False), or kcal/mol (convert_unit=True)

    Parameters
    ----------
    settings: instance of yaml file
    device: list
        list of torch devices

    Returns
    -------
    generator: train, val, test generators, respectively
    int: n_steps for train, val, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data

    """
    from .pyanitools import anidataloader
    atomic_Z_map = {'C': 6, 'H': 1, 'O': 8, 'N': 7}
    # atomic_self_energy = {'H': -0.60467592, 'C': -38.06846167, 'N': -54.70613008, 'O': -75.1796043 } # calculated from dataset
    atomic_self_energy = {'H': -0.500607632585, 'C': -37.8302333826, 'N': -54.5680045287, 'O': -75.0362229210 } # provided by ANI authors
    # meta data
    root = settings['data']['root']
    train_data = settings['data']['train']
    test_data = settings['data']['test']   # can be False

    # Handle train data and test data to make them lists
    if isinstance(train_data, int):
        train_data = [train_data]

    if test_data and isinstance(test_data, int):
        test_data = [test_data]

    test_size_per_molecule = settings['data']['test_size_per_molecule']

    dtrain = {'R':[], 'Z':[], 'N':[], 'E':[]}
    dtest = {'R':[], 'Z':[], 'N':[], 'E':[]}

    for train_data_num in train_data:
        ani_data = anidataloader(root + "/ani_gdb_s0%d.h5" % train_data_num)
        for molecule in ani_data:
            # prepare self energy of the current molecule for subtraction from total energy
            if settings['data']['subtract_self_energy']:
                self_energy = np.sum([atomic_self_energy[a] for a in molecule['species']])
                molecule['energies'] -= self_energy
            if settings['data'].get('convert_unit', True):
                # convert Hartree units to kCal/mol
                molecule['energies'] *= 627.2396
            n_conf, n_atoms, _ = molecule['coordinates'].shape
            conf_indices = np.arange(n_conf)
            # If no test data specified, for each molecule split conformations into train and test
            if test_data is False:
                train_idx, test_idx = train_test_split(conf_indices,
                    test_size=math.ceil(test_size_per_molecule * n_conf),
                    random_state=settings['data']['train_test_split_random_state'])
            else:
                train_idx = conf_indices
            n_conf_train = len(train_idx)
            dtrain['R'].extend(molecule['coordinates'][train_idx])
            dtrain['Z'].extend(np.tile([atomic_Z_map[a] for a in molecule['species']], (n_conf_train, 1)))
            dtrain['N'].extend([n_atoms] * n_conf_train)
            dtrain['E'].extend(molecule['energies'][train_idx])
            if test_data is False:
                n_conf_test = len(test_idx)
                dtest['R'].extend(molecule['coordinates'][test_idx])
                dtest['Z'].extend(np.tile([atomic_Z_map[a] for a in molecule['species']], (n_conf_test, 1)))
                dtest['N'].extend([n_atoms] * n_conf_test)
                dtest['E'].extend(molecule['energies'][test_idx])

    if test_data:
        for test_data_num in test_data:
            ani_data = anidataloader(root + "/ani_gdb_s0%d.h5" % test_data_num)
            for molecule in ani_data:
                 # prepare self energy of the current molecule for subtraction from total energy
                if settings['data']['subtract_self_energy']:
                    self_energy = np.sum([atomic_self_energy[a] for a in molecule['species']])
                    molecule['energies'] -= self_energy
                if settings['data'].get('convert_unit', True):
                    # convert Hartree units to kCal/mol
                    molecule['energies'] *= 627.2396
                n_conf, n_atoms, _ = molecule['coordinates'].shape
                dtest['R'].extend(molecule['coordinates'])
                dtest['Z'].extend(np.tile([atomic_Z_map[a] for a in molecule['species']], (n_conf, 1)))
                dtest['N'].extend([n_atoms] * n_conf)
                dtest['E'].extend(molecule['energies'])

    # Pad irregular-shaped arrays to make all arrays regular in size
    # for k in ['R', 'Z', 'N', 'E']:
    #     dtrain[k] = standardize_batch(dtrain[k])
    #     dtest[k] = standardize_batch(dtest[k])

    further_split_indices = list(range(len(dtrain['R'])))
    train_proportion = settings['data']['train_size_proportion']
    val_proportion = settings['data']['val_size_proportion']
    total_proportion = train_proportion + val_proportion
    if total_proportion < 1:
        train_val_indices, unused_indices = train_test_split(further_split_indices, train_size=total_proportion, random_state=settings['data']['train_val_split_random_state'])
    else:
        train_val_indices = further_split_indices
    train_idx, val_idx = train_test_split(train_val_indices, test_size=(val_proportion / total_proportion), random_state=settings['data']['train_val_split_random_state'])
    data = dtrain
    dtrain = {}
    dval = {}
    for k in data:
        dtrain[k] = np.array(data[k])[train_idx]
        dval[k] = np.array(data[k])[val_idx]
        dtest[k] = np.array(dtest[k])

    # extract data stats
    normalizer = (dtrain['E'].mean(), dtrain['E'].std())

    n_tr_data = len(dtrain['R'])
    n_val_data = len(dval['R'])
    n_test_data = len(dtest['R'])
    print("data size: (train,val,test): %i, %i, %i"%(n_tr_data,n_val_data,n_test_data))

    # HASH check for test energies to make sure test data is fixed
    import hashlib
    test_energy_hash = hashlib.sha1(dtest['E']).hexdigest()
    print("Test set energy HASH:", test_energy_hash)

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # generators
    me = settings['general']['driver']

    # steps
    tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
    val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
    test_steps= int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

    env = ExtensiveEnvironment()

    train_gen = extensive_train_loader(data=dtrain,
                                        env_provider=env,
                                        batch_size=tr_batch_size,
                                        n_rotations=tr_rotations,
                                        freeze_rotations=settings['training']['tr_frz_rot'],
                                        keep_original=settings['training']['tr_keep_original'],
                                        device=device,
                                        shuffle=settings['training']['shuffle'],
                                        drop_last=settings['training']['drop_last'])

    val_gen = extensive_train_loader(data=dval,
                                        env_provider=env,
                                        batch_size=val_batch_size,
                                        n_rotations=val_rotations,
                                        freeze_rotations=settings['training']['val_frz_rot'],
                                        keep_original=settings['training']['val_keep_original'],
                                        device=device,
                                        shuffle=settings['training']['shuffle'],
                                        drop_last=settings['training']['drop_last'])


    test_gen = extensive_train_loader(data=dtest,
                                        env_provider=env,
                                        batch_size=val_batch_size,
                                        n_rotations=val_rotations,
                                        freeze_rotations=settings['training']['val_frz_rot'],
                                        keep_original=settings['training']['val_keep_original'],
                                        device=device,
                                        shuffle=False,
                                        drop_last=False)

    return train_gen, val_gen, test_gen, tr_steps, val_steps, test_steps, n_tr_data, n_val_data, n_test_data, normalizer, test_energy_hash

def parse_t1x_data(settings, device):
    """
    implementation based on train and validation size.
    we don't need the test_size in this implementaion.

    Parameters
    ----------
    settings: instance of yaml file
    device: torch.device
        list of torch devices

    Returns
    -------
    generator: train, val, test generators, respectively
    int: n_steps for train, val, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data

    """
    # meta data
    train_path = settings['data']['train_path']
    val_path = settings['data']['val_path']
    test_path = settings['data']['test_path']   # can be False



    # read data
    dtrain = dict(np.load(train_path))
    dval = dict(np.load(val_path))
    dtest = dict(np.load(test_path))

    # convert unit
    # dtrain['E'] = dtrain['E'] * 23.061
    # dtrain['F'] = dtrain['F'] * 23.061
    # dval['E'] = dval['E'] * 23.061
    # dval['F'] = dval['F'] * 23.061
    # dtest['E'] = dtest['E'] * 23.061
    # dtest['F'] = dtest['F'] * 23.061
    
    # sample data
    if settings['data']['train_size'] != -1:
        n_select = sample_without_replacement(dtrain['R'].shape[0], 
                                              settings['data']['train_size'], 
                                              random_state=settings['data']['random_states'])
        for key in dtrain.keys():
            dtrain[key] = dtrain[key][n_select]
    if settings['data']['val_size'] != -1:
        n_select = sample_without_replacement(dval['R'].shape[0], 
                                              settings['data']['val_size'], 
                                              random_state=settings['data']['random_states'])
        for key in dval.keys():
            dval[key] = dval[key][n_select]
    if settings['data']['test_size'] != -1:
        n_select = sample_without_replacement(dtest['R'].shape[0], 
                                              settings['data']['test_size'], 
                                              random_state=settings['data']['random_states'])
        for key in dtest.keys():
            dtest[key] = dtest[key][n_select]

    # extract data stats
    normalizer = (dtrain['E'].mean(), dtrain['E'].std())

    n_tr_data = dtrain['R'].shape[0]
    n_val_data = dval['R'].shape[0]
    n_test_data = dtest['R'].shape[0]
    print("data size: (train,val,test): %i, %i, %i"%(n_tr_data,n_val_data,n_test_data))

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # generators
    me = settings['general']['driver']

    # steps
    tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
    val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
    test_steps = int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

    env = ExtensiveEnvironment()

    train_gen = extensive_train_loader(data=dtrain,
                                       env_provider=env,
                                       batch_size=tr_batch_size,
                                       n_rotations=tr_rotations,
                                       freeze_rotations=settings['training']['tr_frz_rot'],
                                       keep_original=settings['training']['tr_keep_original'],
                                       device=device,
                                       shuffle=settings['training']['shuffle'],
                                       drop_last=settings['training']['drop_last'])

    val_gen = extensive_train_loader(data=dval,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=settings['training']['shuffle'],
                                     drop_last=settings['training']['drop_last'])

    test_gen = extensive_train_loader(data=dtest,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=False,
                                     drop_last=False)

    return train_gen, val_gen, test_gen, tr_steps, val_steps, test_steps, normalizer

def parse_train_test(settings, device, unit='kcal'):
    """
    implementation based on train and validation size.
    we don't need the test_size in this implementaion.

    Parameters
    ----------
    settings: instance of yaml file
    device: torch.device
        list of torch devices

    Returns
    -------
    generator: train, val, test generators, respectively
    int: n_steps for train, val, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data

    """
    # meta data
    train_path = settings['data']['train_path']
    test_path = settings['data']['test_path']   # can be False

    train_size = settings['data']['train_size']
    val_size = settings['data']['val_size']


    # read data
    data = np.load(train_path)
    test = None
    if test_path:
        test = dict(np.load(test_path))

    # take care of inconsistencies
    dtrain = dict()
    dtest = dict()

    for key in list(data.keys()):
        # copy Z embarrassingly Todo: make it data efficient by flexible environment module
        if key == 'z':
            dtrain['Z'] = np.tile(data['z'], (data['E'].shape[0], 1))
            if test is not None:
                dtest['Z'] = np.tile(test['z'], (test['E'].shape[0], 1))

        elif key == 'E':
            if data['E'].ndim == 1:
                dtrain['E'] = data['E'].reshape(-1,1)
            else:
                dtrain[key] = data[key]

            if test is not None:
                if test['E'].ndim == 1:
                    dtest['E'] = test['E'].reshape(-1, 1)
                else:
                    dtest[key] = test[key]

        elif key in ['R','F','Z']:
            dtrain[key] = data[key]
            if test is not None:
                dtest[key] = test[key]

    # convert unit
    if unit == 'ev':
        dtrain['E'] = dtrain['E'] * 23.061
        dtrain['F'] = dtrain['F'] * 23.061

    # split the data
    dtrain, dval, dtest_leftover = split(dtrain,
                                        train_size=train_size,
                                        test_size=None,
                                        val_size=val_size,
                                        random_states=settings['data']['random_states'])
    if test is None:
        test_size = settings['data'].get('test_size', -1)
        if test_size == -1:
            dtest = dtest_leftover
        else:
            test_size = min(test_size, dtest_leftover['R'].shape[0])
            n_select = sample_without_replacement(dtest_leftover['R'].shape[0],
                                                  test_size,
                                                  random_state=settings['data']['random_states'])
            for k in dtest_leftover.keys():
                dtest[k] = dtest_leftover[k][n_select]


    # extract data stats
    normalizer = (dtrain['E'].mean(), dtrain['E'].std())

    n_tr_data = dtrain['R'].shape[0]
    n_val_data = dval['R'].shape[0]
    n_test_data = dtest['R'].shape[0]
    print("data size: (train,val,test): %i, %i, %i"%(n_tr_data,n_val_data,n_test_data))

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # generators
    me = settings['general']['driver']

    # steps
    tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
    val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
    test_steps= int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

    env = ExtensiveEnvironment()

    train_gen = extensive_train_loader(data=dtrain,
                                       env_provider=env,
                                       batch_size=tr_batch_size,
                                       n_rotations=tr_rotations,
                                       freeze_rotations=settings['training']['tr_frz_rot'],
                                       keep_original=settings['training']['tr_keep_original'],
                                       device=device,
                                       shuffle=settings['training']['shuffle'],
                                       drop_last=settings['training']['drop_last'])

    val_gen = extensive_train_loader(data=dval,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=settings['training']['shuffle'],
                                     drop_last=settings['training']['drop_last'])

    test_gen = extensive_train_loader(data=dtest,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=False,
                                     drop_last=False)

    return train_gen, val_gen, test_gen, tr_steps, val_steps, test_steps, normalizer

def parse_methane_data(settings, device):
    """
    parse and load methane combustion reaction by splitting all available data into train, validation and test

    data comes from https://figshare.com/articles/dataset/Dataset_for_methane_combustion/12973055

    reference paper: Zeng, J., Cao, L., Xu, M. et al. Complex reaction processes in combustion unraveled by neural network-based molecular dynamics simulation. Nat Commun 11, 5713 (2020). 
    https://doi.org/10.1038/s41467-020-19497-z

    energy units: eV ?
    force units: eV/A ?

    Parameters
    ----------
    settings: instance of yaml file
    device: torch.device
        list of torch devices

    Returns
    -------
    generator: train, val, test generators, respectively
    int: n_steps for train, val, test, respectively
    tuple: tuple of mean and standard deviation of energies in the training data

    """
    # meta data
    train_path = settings['data']['train_path']
    test_path = settings['data']['test_path']   # can be False
    assert test_path is False


    all_data = os.listdir(train_path)

    data = {'R':[], 'Z':[], 'E':[], 'F':[], 'NA': []}

    type_maps = {0: 6,  #C,
                 1: 1,  #H,
                 2: 8} #O
    atomic_self_energy = [-1034.54661575, -15.63566308, -2043.59323628] # C, H, O

    for composition in all_data:
        types = np.loadtxt(os.path.join(train_path, composition, "type.raw"), dtype=int)
        Z = np.array([type_maps[i] for i in types])
        n_atoms = len(Z)
        coords = np.load(os.path.join(train_path, composition, "set.000", "coord.npy")).reshape((-1, n_atoms, 3))
        energy = np.load(os.path.join(train_path, composition, "set.000", "energy.npy"))
        force = np.load(os.path.join(train_path, composition, "set.000", "force.npy")).reshape((-1, n_atoms, 3))
        # subtract self energy when needed
        if settings['data']['subtract_self_energy']:
            self_energy = np.sum([atomic_self_energy[i] for i in types])
            energy -= self_energy
        n_confs = energy.shape[0]
        Z = np.tile(Z[None], (n_confs, 1))


        # add compositon data into dataset
        data['R'].extend(coords)
        data['Z'].extend(Z)
        data['E'].extend(energy)
        data['F'].extend(force)
        data['NA'].extend([n_atoms] * n_confs)

    # split data into train and test
    all_indices = list(range(len(data['R'])))
    further_split_indices, test_idx = train_test_split(all_indices, test_size=settings['data']['test_count'], random_state=settings['data']['train_test_split_random_state'])
    train_proportion = settings['data']['train_proportion']
    val_proportion = settings['data']['val_proportion']
    total_proportion = train_proportion + val_proportion
    if total_proportion < 1:
        train_val_indices, unused_indices = train_test_split(further_split_indices, train_size=total_proportion, random_state=settings['data']['train_val_split_random_state'])
    else:
        train_val_indices = further_split_indices
    train_idx, val_idx = train_test_split(train_val_indices, test_size=(val_proportion / total_proportion), random_state=settings['data']['train_val_split_random_state'])

    dtrain = {}
    dval = {}
    dtest = {}
    for k in data:
        dtrain[k] = np.array(data[k])[train_idx]
        dval[k] = np.array(data[k])[val_idx]
        dtest[k] = np.array(data[k])[test_idx]


    # extract data stats
    normalizer = (dtrain['E'].mean(), dtrain['E'].std())

    n_tr_data = dtrain['R'].shape[0]
    n_val_data = dval['R'].shape[0]
    n_test_data = dtest['R'].shape[0]
    print("data size: (train,val,test): %i, %i, %i"%(n_tr_data,n_val_data,n_test_data))

    # HASH check for test energies to make sure test data is fixed
    import hashlib
    test_energy_hash = hashlib.sha1(dtest['E']).hexdigest()
    print("Test set energy HASH:", test_energy_hash)

    tr_batch_size = settings['training']['tr_batch_size']
    val_batch_size = settings['training']['val_batch_size']
    tr_rotations = settings['training']['tr_rotations']
    val_rotations = settings['training']['val_rotations']

    # generators
    me = settings['general']['driver']

    # steps
    tr_steps = int(np.ceil(n_tr_data / tr_batch_size)) * (tr_rotations + 1)
    val_steps = int(np.ceil(n_val_data / val_batch_size)) * (val_rotations + 1)
    test_steps= int(np.ceil(n_test_data / val_batch_size)) * (val_rotations + 1)

    env = ExtensiveEnvironment()

    train_gen = extensive_train_loader(data=dtrain,
                                       env_provider=env,
                                       batch_size=tr_batch_size,
                                       n_rotations=tr_rotations,
                                       freeze_rotations=settings['training']['tr_frz_rot'],
                                       keep_original=settings['training']['tr_keep_original'],
                                       device=device,
                                       shuffle=settings['training']['shuffle'],
                                       drop_last=settings['training']['drop_last'])

    val_gen = extensive_train_loader(data=dval,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=settings['training']['shuffle'],
                                     drop_last=settings['training']['drop_last'])

    test_gen = extensive_train_loader(data=dtest,
                                     env_provider=env,
                                     batch_size=val_batch_size,
                                     n_rotations=val_rotations,
                                     freeze_rotations=settings['training']['val_frz_rot'],
                                     keep_original=settings['training']['val_keep_original'],
                                     device=device,
                                     shuffle=False,
                                     drop_last=False)

    return train_gen, val_gen, test_gen, tr_steps, val_steps, test_steps, n_tr_data, n_val_data, n_test_data, normalizer, test_energy_hash
